# cython: profile=False, boundscheck=False, wraparound=False, cdivision=True
# Profiling is enabled by default as the overhead does not seem to be
# measurable on this specific use case.

# Author: Peter Prettenhofer <peter.prettenhofer@gmail.com>
#         Olivier Grisel <olivier.grisel@ensta.org>
#         Lars Buitinck
#
# License: BSD 3 clause

# TODO: We still need to use ndarrays instead of typed memoryviews when using
# fused types and when the array may be read-only (for instance when it's
# provided by the user). This is fixed in cython > 0.3.

import numpy as np
cimport numpy as np
cimport cython
from cython cimport floating
from libc.math cimport sqrt

try:
    from sklearn.utils.extmath import row_norms
except ImportError:  # pragma: no cover
    row_norms = None


np.import_array()


ctypedef np.float64_t DOUBLE
ctypedef np.int32_t INT


cdef floating _euclidean_dense_dense(
        floating* a,  # IN
        floating* b,  # IN
        int n_features,
        bint squared) nogil:
    """Euclidean distance between a dense and b dense"""
    cdef:
        int i
        int n = n_features // 4
        int rem = n_features % 4
        floating result = 0

    # We manually unroll the loop for better cache optimization.
    for i in range(n):
        result += ((a[0] - b[0]) * (a[0] - b[0])
                  +(a[1] - b[1]) * (a[1] - b[1])
                  +(a[2] - b[2]) * (a[2] - b[2])
                  +(a[3] - b[3]) * (a[3] - b[3]))
        a += 4; b += 4

    for i in range(rem):
        result += (a[i] - b[i]) * (a[i] - b[i])

    return result if squared else sqrt(result)


cdef floating _euclidean_sparse_dense(
        floating[::1] a_data,  # IN
        int[::1] a_indices,    # IN
        floating[::1] b,       # IN
        floating b_squared_norm,
        bint squared) nogil:
    """Euclidean distance between a sparse and b dense"""
    cdef:
        int nnz = a_indices.shape[0]
        int i
        floating tmp, bi
        floating result = 0.0

    for i in range(nnz):
        bi = b[a_indices[i]]
        tmp = a_data[i] - bi
        result += tmp * tmp - bi * bi

    result += b_squared_norm

    if result < 0: result = 0.0

    return result if squared else sqrt(result)


cpdef floating _inertia_dense(
        np.ndarray[floating, ndim=2, mode='c'] X,  # IN
        floating[::1] sample_weight,               # IN
        floating[:, ::1] centers,                  # IN
        int[::1] labels):                          # IN
    """Compute inertia for dense input data

    Sum of squared distance between each sample and its assigned center.
    """
    cdef:
        int n_samples = X.shape[0]
        int n_features = X.shape[1]
        int i, j

        floating sq_dist = 0.0
        floating inertia = 0.0

    for i in range(n_samples):
        j = labels[i]
        sq_dist = _euclidean_dense_dense(&X[i, 0], &centers[j, 0],
                                         n_features, True)
        inertia += sq_dist * sample_weight[i]

    return inertia


cpdef floating _inertia_sparse(
        X,                            # IN
        floating[::1] sample_weight,  # IN
        floating[:, ::1] centers,     # IN
        int[::1] labels):             # IN
    """Compute inertia for sparse input data

    Sum of squared distance between each sample and its assigned center.
    """
    cdef:
        floating[::1] X_data = X.data
        int[::1] X_indices = X.indices
        int[::1] X_indptr = X.indptr

        int n_samples = X.shape[0]
        int n_features = X.shape[1]
        int i, j

        floating sq_dist = 0.0
        floating inertia = 0.0

        floating[::1] centers_squared_norms = row_norms(centers, squared=True)

    for i in range(n_samples):
        j = labels[i]
        sq_dist = _euclidean_sparse_dense(
            X_data[X_indptr[i]: X_indptr[i + 1]],
            X_indices[X_indptr[i]: X_indptr[i + 1]],
            centers[j], centers_squared_norms[j], True)
        inertia += sq_dist * sample_weight[i]

    return inertia


cdef void _average_centers(
        floating[:, ::1] centers,           # INOUT
        floating[::1] weight_in_clusters):  # IN
    """Average new centers wrt weights."""
    cdef:
        int n_clusters = centers.shape[0]
        int n_features = centers.shape[1]
        int j, k
        floating alpha

    for j in range(n_clusters):
        if weight_in_clusters[j] > 0:
            alpha = 1.0 / weight_in_clusters[j]
            for k in range(n_features):
                centers[j, k] *= alpha


cdef void _center_shift(
        floating[:, ::1] centers_old,  # IN
        floating[:, ::1] centers_new,  # IN
        floating[::1] center_shift):   # OUT
    """Compute shift between old and new centers."""
    cdef:
        int n_clusters = centers_old.shape[0]
        int n_features = centers_old.shape[1]
        int j

    for j in range(n_clusters):
        center_shift[j] = _euclidean_dense_dense(
            &centers_new[j, 0], &centers_old[j, 0], n_features, False)


def update_center(
        floating[:, ::1] centers_old,       # IN
        floating[:, ::1] centers_new,       # INOUT
        floating[::1] center_shift,         # OUT
        floating[::1] weight_in_clusters):  # IN
    _average_centers(centers_new, weight_in_clusters)
    _center_shift(centers_old, centers_new, center_shift)


def merge_update_chunks(int n_clusters,
                        int n_features,
                        floating[::1] weight_in_clusters,
                        floating[::1] weight_in_clusters_chunk,
                        floating[:, ::1] centers_new,
                        floating[:, ::1] centers_new_chunk):
    for j in range(n_clusters):
        weight_in_clusters[j] += weight_in_clusters_chunk[j]
        for k in range(n_features):
            centers_new[j, k] += centers_new_chunk[j, k]


def update_upper_lower_bounds(
        floating[::1] upper_bounds,                # INOUT
        floating[:, ::1] lower_bounds,             # INOUT
        int[::1] labels,                           # IN
        floating[::1] center_shift):               # IN
    cdef:
        int n_samples = upper_bounds.shape[0]
        int n_clusters = lower_bounds.shape[1]

    for i in range(n_samples):
        upper_bounds[i] += center_shift[labels[i]]

        for j in range(n_clusters):
            lower_bounds[i, j] -= center_shift[j]
            if lower_bounds[i, j] < 0:
                lower_bounds[i, j] = 0



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import tempfile
import time
from typing import Dict, Iterator, Optional, List, Set

from ....core import ChunkGraph, TileableGraph, Chunk, TileContext
from ....core.operand import Fetch
from ....metrics import Metrics
from ....optimization.logical import OptimizationRecords
from ....oscar.profiling import (
    ProfilingData,
    MARS_ENABLE_PROFILING,
)
from ....typing import TileableType, ChunkType
from ....utils import Timer
from ...subtask import SubtaskResult, Subtask
from ..core import Task, TaskResult, TaskStatus, new_task_id
from ..execution.api import TaskExecutor, ExecutionChunkResult
from .preprocessor import TaskPreprocessor

logger = logging.getLogger(__name__)

MARS_ENABLE_DUMPING_SUBTASK_GRAPH = int(os.environ.get("MARS_DUMP_SUBTASK_GRAPH", 0))


class TaskProcessor:
    _tileable_to_subtasks: Dict[TileableType, List[Subtask]]
    _tileable_id_to_tileable: Dict[str, TileableType]
    _chunk_to_subtasks: Dict[ChunkType, Subtask]
    _stage_tileables: Set[TileableType]

    def __init__(
        self,
        task: Task,
        preprocessor: TaskPreprocessor,
        executor: TaskExecutor,
    ):
        self._task = task
        self._preprocessor = preprocessor
        self._executor = executor

        self._tileable_id_to_tileable = dict()
        self._chunk_to_subtasks = dict()
        self._stage_tileables = set()

        if MARS_ENABLE_PROFILING:
            ProfilingData.init(task.task_id)
        elif task.extra_config and task.extra_config.get("enable_profiling"):
            ProfilingData.init(task.task_id, task.extra_config["enable_profiling"])

        self._dump_subtask_graph = False
        if MARS_ENABLE_DUMPING_SUBTASK_GRAPH or (
            task.extra_config and task.extra_config.get("dump_subtask_graph")
        ):
            self._dump_subtask_graph = True

        self.result = TaskResult(
            task_id=task.task_id,
            session_id=task.session_id,
            start_time=time.time(),
            status=TaskStatus.pending,
        )
        self.done = asyncio.Event()

        # add metrics
        self._chunk_graph_gen_time = Metrics.gauge(
            "mars.chunk_graph_gen_time_secs",
            "Time consuming in seconds to generate a chunk graph",
            ("session_id", "task_id"),
        )
        self._subtask_graph_gen_time = Metrics.gauge(
            "mars.subtask_graph_gen_time_secs",
            "Time consuming in seconds to generate a subtask graph",
            ("session_id", "task_id", "stage_id"),
        )
        self._task_execution_time = Metrics.gauge(
            "mars.task_execution_time_secs",
            "Time consuming in seconds to execute a task",
            ("session_id", "task_id"),
        )

    @property
    def task_id(self):
        return self._task.task_id

    @property
    def tileable_graph(self):
        return self._preprocessor.tileable_graph

    @property
    def tileable_id_to_tileable(self):
        return self._tileable_id_to_tileable

    @property
    def tile_context(self) -> TileContext:
        return self._preprocessor.tile_context

    @property
    def stage_processors(self):
        # TODO(fyrestone): Remove it.
        return self._executor.get_stage_processors()

    def get_tiled(self, tileable: TileableType):
        return self._preprocessor.get_tiled(tileable)

    def get_subtasks(self, chunks: List[ChunkType]) -> List[Subtask]:
        return [self._chunk_to_subtasks[chunk] for chunk in chunks]

    def get_tileable_to_subtasks(self) -> Dict[TileableType, List[Subtask]]:
        tile_context = self.tile_context
        result = dict()
        for tileable, infos in tile_context.get_tileable_tile_infos().items():
            subtasks = []
            for info in infos:
                chunks = [
                    c for c in info.generated_chunks if not isinstance(c.op, Fetch)
                ]
                subtasks.extend(self.get_subtasks(chunks))
            result[tileable] = subtasks
        return result

    @staticmethod
    async def _get_next_chunk_graph(
        chunk_graph_iter: Iterator[ChunkGraph],
    ) -> Optional[ChunkGraph]:
        def next_chunk_graph():
            try:
                return next(chunk_graph_iter)
            except StopIteration:
                return

        fut = asyncio.to_thread(next_chunk_graph)
        chunk_graph = await fut
        return chunk_graph

    async def _iter_stage_chunk_graph(self):
        tileable_graph = self._preprocessor.tileable_graph
        chunk_graph_iter = iter(self._preprocessor.tile(tileable_graph))
        while True:
            with Timer() as stage_timer:
                with Timer() as timer:
                    chunk_graph = await self._get_next_chunk_graph(chunk_graph_iter)
                    if chunk_graph is None:
                        # tile finished
                        self._preprocessor.done = True
                        return
                stage_id = new_task_id()
                stage_profiler = ProfilingData[self._task.task_id, "general"].nest(
                    f"stage_{stage_id}"
                )
                stage_profiler.set(f"tile({len(chunk_graph)})", timer.duration)
                logger.info(
                    "Time consuming to gen a chunk graph is %ss with session id %s, task id %s",
                    timer.duration,
                    self._task.session_id,
                    self._task.task_id,
                )
                self._chunk_graph_gen_time.record(
                    timer.duration,
                    {
                        "session_id": self._task.session_id,
                        "task_id": self._task.task_id,
                    },
                )
                yield stage_id, stage_profiler, chunk_graph

            stage_profiler.set("total", stage_timer.duration)

    async def _process_stage_chunk_graph(
        self,
        stage_id: str,
        stage_profiler,
        chunk_graph: ChunkGraph,
    ):
        available_bands = await self._executor.get_available_band_resources()
        meta_api = self._executor._meta_api
        get_meta_tasks = []
        fetch_op_keys = []
        for c in chunk_graph.iter_indep():
            if isinstance(c.op, Fetch):
                get_meta_tasks.append(
                    meta_api.get_chunk_meta.delay(c.key, fields=["bands"])
                )
                fetch_op_keys.append(c.op.key)
        # TODO(fyrestone): A more general way to get the key to bands
        # for all execution backends.
        try:
            key_to_bands = await meta_api.get_chunk_meta.batch(*get_meta_tasks)
            fetch_op_to_bands = dict(
                (key, meta["bands"][0])
                for key, meta in zip(fetch_op_keys, key_to_bands)
            )
        except (KeyError, IndexError):
            fetch_op_to_bands = {}
        with Timer() as timer:
            subtask_graph = await asyncio.to_thread(
                self._preprocessor.analyze,
                chunk_graph,
                self._chunk_to_subtasks,
                available_bands,
                stage_id=stage_id,
                op_to_bands=fetch_op_to_bands,
            )
        stage_profiler.set(f"gen_subtask_graph({len(subtask_graph)})", timer.duration)
        logger.info(
            "Time consuming to gen a subtask graph is %ss with session id %s, task id %s, stage id %s",
            timer.duration,
            self._task.session_id,
            self._task.task_id,
            stage_id,
        )
        self._subtask_graph_gen_time.record(
            timer.duration,
            {
                "session_id": self._task.session_id,
                "task_id": self._task.task_id,
                "stage_id": stage_id,
            },
        )

        tile_context = await asyncio.to_thread(
            self._get_stage_tile_context,
            {c for c in chunk_graph.result_chunks if not isinstance(c.op, Fetch)},
        )

        with Timer() as timer:
            chunk_to_result = await self._executor.execute_subtask_graph(
                stage_id, subtask_graph, chunk_graph, tile_context
            )
        stage_profiler.set("run", timer.duration)

        self._preprocessor.post_chunk_graph_execution()
        if self._preprocessor.chunk_optimization_records_list:
            optimization_records = self._preprocessor.chunk_optimization_records_list[
                -1
            ]
        else:
            optimization_records = None
        self._update_stage_meta(chunk_to_result, tile_context, optimization_records)

    def _get_stage_tile_context(self, result_chunks: Set[Chunk]) -> TileContext:
        collected = self._stage_tileables
        tile_context = TileContext()
        for tileable in self.tileable_graph:
            if tileable in collected:
                continue
            tiled_tileable = self._preprocessor.tile_context.get(tileable)
            if tiled_tileable is not None:
                tileable_chunks = [c.data for c in tiled_tileable.chunks]
                if any(c not in result_chunks for c in tileable_chunks):
                    continue
                tile_context[tileable] = tiled_tileable
                collected.add(tileable)
        return tile_context

    @classmethod
    def _update_stage_meta(
        cls,
        chunk_to_result: Dict[Chunk, ExecutionChunkResult],
        tile_context: TileContext,
        optimization_records: OptimizationRecords,
    ):
        for tiled_tileable in tile_context.values():
            cls._update_result_meta(chunk_to_result, tiled_tileable)

        for c, r in chunk_to_result.items():
            c.params = r.meta
            original_chunk = (
                optimization_records and optimization_records.get_original_entity(c)
            )
            if original_chunk is not None:
                original_chunk.params = r.meta

        for tileable, tiled_tileable in tile_context.items():
            tiled_tileable.refresh_params()
            tileable.params = tiled_tileable.params

    @classmethod
    def _update_result_meta(
        cls, chunk_to_result: Dict[Chunk, ExecutionChunkResult], tileable: TileableType
    ):
        from ....dataframe.core import DATAFRAME_TYPE, SERIES_TYPE
        from ....tensor.core import TENSOR_TYPE

        chunks = [c.data for c in tileable.chunks]
        if isinstance(tileable, DATAFRAME_TYPE):
            for c in chunks:
                i, j = c.index
                meta = chunk_to_result[c].meta
                shape = meta.get("shape")
                update_shape = shape is None
                shape = shape if not update_shape else [None, None]
                if i > 0:
                    # update dtypes_value
                    c0j = chunk_to_result[tileable.cix[0, j].data].meta
                    meta["dtypes_value"] = c0j["dtypes_value"]
                    if update_shape:
                        shape[1] = c0j["shape"][1]
                if j > 0:
                    # update index_value
                    ci0 = chunk_to_result[tileable.cix[i, 0].data].meta
                    meta["index_value"] = ci0["index_value"]
                    if update_shape:
                        shape[0] = ci0["shape"][0]
                if update_shape:
                    meta["shape"] = tuple(shape)
        elif isinstance(tileable, SERIES_TYPE):
            first_meta = chunk_to_result[chunks[0]].meta
            for c in chunks:
                i = c.index[0]
                meta = chunk_to_result[c].meta
                if i > 0:
                    meta["name"] = first_meta["name"]
                    meta["dtype"] = first_meta["dtype"]
        elif isinstance(tileable, TENSOR_TYPE):
            ndim = tileable.ndim
            for i, c in enumerate(chunks):
                meta = chunk_to_result[c].meta
                if "shape" not in meta:
                    shape = []
                    for i, ind in enumerate(c.index):
                        ind0 = [0] * ndim
                        ind0[i] = ind
                        c0 = tileable.cix[tuple(ind0)].data
                        shape.append(chunk_to_result[c0].meta["shape"][i])
                    meta["shape"] = tuple(shape)
                if i > 0:
                    first = chunk_to_result[chunks[0]].meta
                    meta["dtype"] = first["dtype"]
                    meta["order"] = first["order"]

    async def run(self):
        profiling = ProfilingData[self.task_id, "general"]
        self.result.status = TaskStatus.running
        # optimization
        with Timer() as timer:
            # optimization, run it in executor,
            # since optimization may be a CPU intensive operation
            await asyncio.to_thread(self._preprocessor.optimize)
        profiling.set("optimize", timer.duration)

        self._tileable_id_to_tileable = await asyncio.to_thread(
            self._get_tileable_id_to_tileable, self._preprocessor.tileable_graph
        )

        try:
            async with self._executor:
                async for stage_args in self._iter_stage_chunk_graph():
                    await self._process_stage_chunk_graph(*stage_args)
        except Exception as ex:
            self.result.error = ex
            self.result.traceback = ex.__traceback__
        finally:
            self._gen_result()
            self._finish()

    async def get_progress(self) -> float:
        # get tileable proportion that is tiled
        return await self._executor.get_progress()

    async def cancel(self):
        self._preprocessor.cancel()
        await self._executor.cancel()

    async def set_subtask_result(self, subtask_result: SubtaskResult):
        await self._executor.set_subtask_result(subtask_result)

    @staticmethod
    def _get_tileable_id_to_tileable(
        tileable_graph: TileableGraph,
    ) -> Dict[str, TileableType]:
        tileable_id_to_tileable = dict()

        for tileable in tileable_graph:
            tileable_id_to_tileable[str(tileable.key)] = tileable

        return tileable_id_to_tileable

    def _gen_result(self):
        self.result.status = TaskStatus.terminated
        self.result.end_time = time.time()
        cost_time_secs = self.result.end_time - self.result.start_time
        logger.info(
            "Time consuming to execute a task is %ss with session id %s, task id %s",
            cost_time_secs,
            self._task.session_id,
            self._task.task_id,
        )
        self._task_execution_time.record(
            cost_time_secs,
            {"session_id": self._task.session_id, "task_id": self._task.task_id},
        )

    def dump_subtask_graph(self):
        from .graph_visualizer import GraphVisualizer

        try:  # pragma: no cover
            import graphviz
        except ImportError:
            graphviz = None

        dot = GraphVisualizer(self).to_dot()
        directory = tempfile.gettempdir()
        file_name = f"mars-{self.task_id}"
        logger.debug(
            "subtask graph is stored in %s", os.path.join(directory, file_name)
        )
        if graphviz is not None:  # pragma: no cover
            g = graphviz.Source(dot)
            g.view(file_name, directory=directory)
        else:
            with open(os.path.join(directory, file_name), "w") as f:
                f.write(dot)

    def _finish(self):
        self._executor.destroy()
        self.done.set()
        if self._dump_subtask_graph:
            self.dump_subtask_graph()
        if MARS_ENABLE_PROFILING or (
            self._task.extra_config and self._task.extra_config.get("enable_profiling")
        ):
            ProfilingData[self._task.task_id, "general"].set(
                "total", time.time() - self.result.start_time
            )
            serialization = ProfilingData[self._task.task_id, "serialization"]
            if not serialization.empty():
                serialization.set(
                    "total",
                    sum(serialization.values()),
                )
            data = ProfilingData.pop(self._task.task_id)
            self.result.profiling = {
                "supervisor": data,
            }

    def is_done(self) -> bool:
        return self.done.is_set()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cloudpickle
import numpy as np

from ...core import Object, OBJECT_TYPE, OBJECT_CHUNK_TYPE, recursive_tile
from ...serialization.serializables import (
    KeyField,
    Int32Field,
    DictField,
    AnyField,
    BoolField,
)
from ...tensor.core import TensorOrder
from ...utils import has_unknown_shape, tokenize
from ..operands import LearnOperand, LearnOperandMixin, OutputType


class TreeObject(Object):
    def fetch(self, session=None, **kw):
        result = self._data.fetch(session=session, **kw)
        return cloudpickle.loads(result) if isinstance(result, bytes) else result


class TreeBase(LearnOperand, LearnOperandMixin):
    _input = KeyField("input")
    _leaf_size = Int32Field("leaf_size")
    _metric = AnyField("metric")

    _metric_params = DictField("metric_params")

    def __init__(
        self, leaf_size=None, metric=None, metric_params=None, output_types=None, **kw
    ):
        super().__init__(
            _leaf_size=leaf_size,
            _metric=metric,
            _metric_params=metric_params,
            _output_types=output_types,
            **kw
        )
        if self.output_types is None:
            self.output_types = [OutputType.object]

    @property
    def input(self):
        return self._input

    @property
    def leaf_size(self):
        return self._leaf_size

    @property
    def metric(self):
        return self._metric

    @property
    def metric_params(self):
        return self._metric_params

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        return self.new_tileable([a])

    @classmethod
    def tile(cls, op):
        if has_unknown_shape(*op.inputs):
            yield

        # ball tree and kd tree requires the full data,
        # thus rechunk input tensor into 1 chunk
        inp = op.input.rechunk({ax: s for ax, s in enumerate(op.input.shape)})
        inp = yield from recursive_tile(inp)
        out = op.outputs[0]

        chunk_op = op.copy().reset_key()
        kw = out.params
        kw["index"] = inp.chunks[0].index
        chunk = chunk_op.new_chunk([inp.chunks[0]], kws=[kw])

        new_op = op.copy()
        tileable_kw = out.params
        tileable_kw["nsplits"] = ((1,),)
        tileable_kw["chunks"] = [chunk]
        return new_op.new_tileables(op.inputs, kws=[tileable_kw])

    @classmethod
    def execute(cls, ctx, op):
        if op.gpu:  # pragma: no cover
            raise NotImplementedError(
                "Does not support tree-based nearest neighbors on GPU"
            )

        a = ctx[op.input.key]
        tree = cls._tree_type(
            a, op.leaf_size, metric=op.metric, **(op.metric_params or dict())
        )
        ctx[op.outputs[0].key] = tree


def _on_serialize_tree(tree):
    return cloudpickle.dumps(tree) if not hasattr(tree, "key") else tree


def _on_deserialize_tree(ser):
    return cloudpickle.loads(ser) if isinstance(ser, bytes) else ser


class TreeQueryBase(LearnOperand, LearnOperandMixin):
    _input = KeyField("input")
    _tree = AnyField(
        "tree", on_serialize=_on_serialize_tree, on_deserialize=_on_deserialize_tree
    )
    _n_neighbors = Int32Field("n_neighbors")
    _return_distance = BoolField("return_distance")

    def __init__(
        self, tree=None, n_neighbors=None, return_distance=None, output_types=None, **kw
    ):
        super().__init__(
            _tree=tree,
            _n_neighbors=n_neighbors,
            _return_distance=return_distance,
            _output_types=output_types,
            **kw
        )
        if self.output_types is None:
            self.output_types = [OutputType.tensor] * self.output_limit

    @property
    def input(self):
        return self._input

    @property
    def tree(self):
        return self._tree

    @property
    def n_neighbors(self):
        return self._n_neighbors

    @property
    def return_distance(self):
        return self._return_distance

    @property
    def output_limit(self):
        return 2 if self._return_distance else 1

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if isinstance(self._tree, (OBJECT_TYPE, OBJECT_CHUNK_TYPE)):
            self._tree = self._inputs[1]

    def _update_key(self):
        values = []
        for value in self._values_:
            if isinstance(value, self._tree_type):
                values.append(cloudpickle.dumps(value))
            else:
                values.append(value)
        self._obj_set("_key", tokenize(type(self).__name__, *values))
        return self

    def __call__(self, x):
        kws = []
        if self._return_distance:
            kws.append(
                {
                    "shape": (x.shape[0], self._n_neighbors),
                    "dtype": np.dtype(np.float64),
                    "order": x.order,
                    "type": "distance",
                }
            )
        kws.append(
            {
                "shape": (x.shape[0], self._n_neighbors),
                "dtype": np.dtype(np.int64),
                "order": TensorOrder.C_ORDER,
                "type": "indices",
            }
        )
        inputs = [x]
        if isinstance(self._tree, OBJECT_TYPE):
            inputs.append(self._tree)
        return self.new_tileables(inputs, kws=kws, output_limit=len(kws))

    @classmethod
    def tile(cls, op):
        inp = op.input

        if inp.chunk_shape[1] != 1:
            if has_unknown_shape(inp):
                yield
            inp = yield from recursive_tile(inp.rechunk({1: inp.shape[1]}))

        tree_chunk = None
        if isinstance(op.tree, OBJECT_TYPE):
            tree_chunk = op.tree.chunks[0]
        out_chunks = [[] for _ in range(len(op.outputs))]
        for chunk in inp.chunks:
            chunk_op = op.copy().reset_key()
            if tree_chunk is not None:
                chunk_op._tree = tree_chunk
            chunk_kws = []
            if op.return_distance:
                chunk_kws.append(
                    {
                        "shape": (chunk.shape[0], op.n_neighbors),
                        "dtype": np.dtype(np.float64),
                        "order": chunk.order,
                        "index": chunk.index,
                        "type": "distance",
                    }
                )
            chunk_kws.append(
                {
                    "shape": (chunk.shape[0], op.n_neighbors),
                    "dtype": np.dtype(np.int64),
                    "order": TensorOrder.C_ORDER,
                    "index": chunk.index,
                    "type": "indices",
                }
            )
            chunk_inputs = [chunk]
            if tree_chunk is not None:
                chunk_inputs.append(tree_chunk)
            chunks = chunk_op.new_chunks(
                chunk_inputs, kws=chunk_kws, output_limit=len(chunk_kws)
            )
            for cs, c in zip(out_chunks, chunks):
                cs.append(c)

        kws = [o.params for o in op.outputs]
        nsplits = list(inp.nsplits)
        nsplits[1] = (op.n_neighbors,)
        if op.return_distance:
            kws[0]["chunks"] = out_chunks[0]
            kws[0]["nsplits"] = tuple(nsplits)
        kws[-1]["chunks"] = out_chunks[-1]
        kws[-1]["nsplits"] = tuple(nsplits)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=kws, output_limit=len(kws))

    @classmethod
    def execute(cls, ctx, op):
        if op.gpu:  # pragma: no cover
            raise NotImplementedError(
                "Does not support tree-based nearest neighbors on GPU"
            )

        x = ctx[op.input.key]
        if len(op.inputs) == 2:
            tree = ctx[op.tree.key]
        else:
            tree = op.tree
        tree = cloudpickle.loads(tree) if isinstance(tree, bytes) else tree
        ret = tree.query(x, op.n_neighbors, op.return_distance)
        if op.return_distance:
            ctx[op.outputs[0].key] = ret[0]
            ctx[op.outputs[1].key] = ret[1]
        else:
            ctx[op.outputs[0].key] = ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number

import numpy as np

from ..datasource import tensor as astensor
from .. import arithmetic as arith, reduction


class TensorUfuncDef:
    def __init__(
        self, method, aggregator=None, accumulator=None, pre_agg=None, post_agg=None
    ):
        self._method = method
        self._aggregator = aggregator
        self._accumulator = accumulator
        self._pre_agg = pre_agg
        self._post_agg = post_agg

    def __call__(self, *args, **kwargs):
        return self._method(*args, **kwargs)

    def at(self, a, indices, b=None):
        # todo handle setting duplicated keys, a separate operand may be needed
        if b is None:
            a[indices] = self(a[indices])
        else:
            a[indices] = self(a[indices], b)

    def accumulate(self, array, axis=0, dtype=None, out=None):
        if self._accumulator is None:
            raise NotImplementedError
        data = array if self._pre_agg is None else self._pre_agg(array)
        result = self._accumulator(data, axis=axis, dtype=dtype)
        result = result if self._post_agg is None else self._post_agg(result)
        if out is not None:
            out[0]._data = result._data
        else:
            return result

    def reduce(self, array, axis=0, dtype=None, out=None, keepdims=False):
        if self._aggregator is None:
            raise NotImplementedError
        data = array if self._pre_agg is None else self._pre_agg(array)
        result = self._aggregator(data, axis=axis, dtype=dtype, keepdims=keepdims)
        result = result if self._post_agg is None else self._post_agg(result)
        if out is not None:
            out[0]._data = result._data
        else:
            return result


UFUNC_TO_TENSOR_FUNCS = {
    np.add: TensorUfuncDef(
        arith.add,
        accumulator=reduction.cumsum,
        aggregator=reduction.sum,
    ),
    np.subtract: TensorUfuncDef(arith.subtract),
    np.multiply: TensorUfuncDef(
        arith.multiply,
        accumulator=reduction.cumprod,
        aggregator=reduction.prod,
    ),
    np.divide: TensorUfuncDef(arith.divide),
    np.logaddexp: TensorUfuncDef(
        arith.logaddexp,
        accumulator=reduction.cumsum,
        aggregator=reduction.sum,
        pre_agg=arith.exp,
        post_agg=arith.log,
    ),
    np.logaddexp2: TensorUfuncDef(
        arith.logaddexp2,
        accumulator=reduction.cumsum,
        aggregator=reduction.sum,
        pre_agg=lambda x: arith.power(2, x),
        post_agg=arith.log2,
    ),
    np.true_divide: TensorUfuncDef(arith.truediv),
    np.floor_divide: TensorUfuncDef(arith.floordiv),
    # unary
    np.negative: TensorUfuncDef(arith.negative),
    np.power: TensorUfuncDef(arith.power),
    np.float_power: TensorUfuncDef(arith.float_power),
    np.remainder: TensorUfuncDef(arith.remainder),
    np.mod: TensorUfuncDef(arith.mod),
    np.fmod: TensorUfuncDef(arith.fmod),
    np.conj: TensorUfuncDef(arith.conj),
    np.conjugate: TensorUfuncDef(arith.conjugate),
    np.exp: TensorUfuncDef(arith.exp),
    np.exp2: TensorUfuncDef(arith.exp2),
    np.log: TensorUfuncDef(arith.log),
    np.log2: TensorUfuncDef(arith.log2),
    np.log10: TensorUfuncDef(arith.log10),
    np.log1p: TensorUfuncDef(arith.log1p),
    np.expm1: TensorUfuncDef(arith.expm1),
    np.sqrt: TensorUfuncDef(arith.sqrt),
    np.square: TensorUfuncDef(arith.square),
    np.cbrt: TensorUfuncDef(arith.cbrt),
    np.reciprocal: TensorUfuncDef(arith.reciprocal),
    # trigonometric functions
    np.sin: TensorUfuncDef(arith.sin),
    np.cos: TensorUfuncDef(arith.cos),
    np.tan: TensorUfuncDef(arith.tan),
    np.arcsin: TensorUfuncDef(arith.arcsin),
    np.arccos: TensorUfuncDef(arith.arccos),
    np.arctan: TensorUfuncDef(arith.arctan),
    np.arctan2: TensorUfuncDef(arith.arctan2),
    np.hypot: TensorUfuncDef(arith.hypot),
    np.sinh: TensorUfuncDef(arith.sinh),
    np.cosh: TensorUfuncDef(arith.cosh),
    np.tanh: TensorUfuncDef(arith.tanh),
    np.arcsinh: TensorUfuncDef(arith.arcsinh),
    np.arccosh: TensorUfuncDef(arith.arccosh),
    np.arctanh: TensorUfuncDef(arith.arctanh),
    np.deg2rad: TensorUfuncDef(arith.deg2rad),
    np.rad2deg: TensorUfuncDef(arith.rad2deg),
    # comparison functions
    np.greater: TensorUfuncDef(arith.greater),
    np.greater_equal: TensorUfuncDef(arith.greater_equal),
    np.less: TensorUfuncDef(arith.less),
    np.less_equal: TensorUfuncDef(arith.less_equal),
    np.not_equal: TensorUfuncDef(arith.not_equal),
    np.equal: TensorUfuncDef(arith.equal),
    np.logical_and: TensorUfuncDef(arith.logical_and),
    np.logical_or: TensorUfuncDef(arith.logical_or),
    np.logical_xor: TensorUfuncDef(arith.logical_xor),
    np.logical_not: TensorUfuncDef(arith.logical_not),
    np.maximum: TensorUfuncDef(arith.maximum),
    np.minimum: TensorUfuncDef(arith.minimum),
    np.fmax: TensorUfuncDef(arith.fmax),
    np.fmin: TensorUfuncDef(arith.fmin),
    # floating functions
    np.isfinite: TensorUfuncDef(arith.isfinite),
    np.isinf: TensorUfuncDef(arith.isinf),
    np.isnan: TensorUfuncDef(arith.isnan),
    np.signbit: TensorUfuncDef(arith.signbit),
    np.copysign: TensorUfuncDef(arith.copysign),
    np.nextafter: TensorUfuncDef(arith.nextafter),
    np.spacing: TensorUfuncDef(arith.spacing),
    np.modf: TensorUfuncDef(arith.modf),
    np.ldexp: TensorUfuncDef(arith.ldexp),
    np.frexp: TensorUfuncDef(arith.frexp),
    np.floor: TensorUfuncDef(arith.floor),
    np.ceil: TensorUfuncDef(arith.ceil),
    np.trunc: TensorUfuncDef(arith.trunc),
    # more math functions
    np.degrees: TensorUfuncDef(arith.degrees),
    np.radians: TensorUfuncDef(arith.radians),
    np.rint: TensorUfuncDef(arith.rint),
    np.fabs: TensorUfuncDef(arith.fabs),
    np.sign: TensorUfuncDef(arith.sign),
    np.absolute: TensorUfuncDef(arith.absolute),
}


def _check_arg(arg):
    if isinstance(arg, Number):
        return True

    try:
        astensor(arg)
        return True
    except ValueError:
        return False


def _array_ufunc(_, ufunc, method, *inputs, **kwargs):
    out = kwargs.get("out", tuple())
    for x in inputs + out:
        if not _check_arg(x):
            return NotImplemented

    if ufunc.signature is not None:
        return NotImplemented
    if ufunc not in UFUNC_TO_TENSOR_FUNCS:
        return NotImplemented

    try:
        tensor_func = getattr(UFUNC_TO_TENSOR_FUNCS[ufunc], method)
        return tensor_func(*inputs, **kwargs)
    except (AttributeError, NotImplementedError):
        return NotImplemented



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...core import ExecutableTuple
from ...serialization.serializables import Int64Field, BoolField, AnyField
from ...config import options
from ..utils import decide_chunk_sizes
from .core import TensorNoInput
from ..array_utils import create_array


class TensorLinspace(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_LINSPACE

    _start = AnyField("start")
    _stop = AnyField("stop")
    _num = Int64Field("num")
    _endpoint = BoolField("endpoint")

    def __init__(
        self, start=None, stop=None, num=None, endpoint=None, dtype=None, **kw
    ):
        dtype = np.dtype(np.linspace(0, 1, 1).dtype if dtype is None else dtype)
        super().__init__(
            _start=start, _stop=stop, _num=num, _endpoint=endpoint, dtype=dtype, **kw
        )

    def to_chunk_op(self, *args):
        start, stop, num, endpoint = args
        op = self.copy().reset_key()
        op._start = start
        op._stop = stop
        op._num = num
        op._endpoint = endpoint
        return op

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        chunk_length = tensor.extra_params.raw_chunk_size or options.chunk_size
        chunk_length = decide_chunk_sizes(
            tensor.shape, chunk_length, tensor.dtype.itemsize
        )

        start, stop, num, endpoint = (
            tensor.op.start,
            tensor.op.stop,
            tensor.op.num,
            tensor.op.endpoint,
        )
        if num > 1:
            step = float(stop - start) / (num if not endpoint else num - 1)
        else:
            step = 0.0

        chunks = []
        chunk_start = start
        nsplit = []
        for i, cs in enumerate(chunk_length[0]):
            chunk_stop = chunk_start + (cs - 1) * step
            chunk_op = op.to_chunk_op(chunk_start, chunk_stop, cs, True)
            chunk_shape = (cs,)
            chunk_idx = (i,)
            chunk = chunk_op.new_chunk(None, shape=chunk_shape, index=chunk_idx)
            chunks.append(chunk)
            nsplit.append(cs)
            chunk_start = chunk_start + cs * step

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs, op.outputs[0].shape, chunks=chunks, nsplits=(tuple(nsplit),)
        )

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def num(self):
        return self._num

    @property
    def endpoint(self):
        return self._endpoint

    @classmethod
    def execute(cls, ctx, op):
        ctx[op.outputs[0].key] = create_array(op)(
            "linspace",
            op.start,
            op.stop,
            num=op.num,
            endpoint=op.endpoint,
            dtype=op.dtype,
        )


def linspace(
    start,
    stop,
    num=50,
    endpoint=True,
    retstep=False,
    dtype=None,
    gpu=None,
    chunk_size=None,
):
    """
    Return evenly spaced numbers over a specified interval.

    Returns `num` evenly spaced samples, calculated over the
    interval [`start`, `stop`].

    The endpoint of the interval can optionally be excluded.

    Parameters
    ----------
    start : scalar
        The starting value of the sequence.
    stop : scalar
        The end value of the sequence, unless `endpoint` is set to False.
        In that case, the sequence consists of all but the last of ``num + 1``
        evenly spaced samples, so that `stop` is excluded.  Note that the step
        size changes when `endpoint` is False.
    num : int, optional
        Number of samples to generate. Default is 50. Must be non-negative.
    endpoint : bool, optional
        If True, `stop` is the last sample. Otherwise, it is not included.
        Default is True.
    retstep : bool, optional
        If True, return (`samples`, `step`), where `step` is the spacing
        between samples.
    dtype : dtype, optional
        The type of the output tensor.  If `dtype` is not given, infer the data
        type from the other input arguments.
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension

    Returns
    -------
    samples : Tensor
        There are `num` equally spaced samples in the closed interval
        ``[start, stop]`` or the half-open interval ``[start, stop)``
        (depending on whether `endpoint` is True or False).
    step : float, optional
        Only returned if `retstep` is True

        Size of spacing between samples.


    See Also
    --------
    arange : Similar to `linspace`, but uses a step size (instead of the
             number of samples).
    logspace : Samples uniformly distributed in log space.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.linspace(2.0, 3.0, num=5).execute()
    array([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ])
    >>> mt.linspace(2.0, 3.0, num=5, endpoint=False).execute()
    array([ 2. ,  2.2,  2.4,  2.6,  2.8])
    >>> mt.linspace(2.0, 3.0, num=5, retstep=True).execute()
    (array([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ]), 0.25)

    Graphical illustration:

    >>> import matplotlib.pyplot as plt
    >>> N = 8
    >>> y = mt.zeros(N)
    >>> x1 = mt.linspace(0, 10, N, endpoint=True)
    >>> x2 = mt.linspace(0, 10, N, endpoint=False)
    >>> plt.plot(x1.execute(), y.execute(), 'o')
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> plt.plot(x2.execute(), y.execute() + 0.5, 'o')
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> plt.ylim([-0.5, 1])
    (-0.5, 1)
    >>> plt.show()

    """
    num = int(num)

    op = TensorLinspace(start, stop, num, endpoint, dtype=dtype, gpu=gpu)
    shape = (num,)
    ret = op(shape, chunk_size=chunk_size)

    if not retstep:
        return ret

    if num > 1:
        step = float(stop - start) / (num if not endpoint else num - 1)
    else:
        step = np.nan

    return ExecutableTuple([ret, step])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...config import options
from ...core import OutputType, ENTITY_TYPE
from ...serialization.serializables import BoolField
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameSkew(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = opcodes.SKEW
    _func_name = "skew"

    _bias = BoolField("bias")

    def __init__(self, bias=None, **kw):
        super().__init__(_bias=bias, **kw)

    @property
    def bias(self):
        return self._bias

    @classmethod
    def get_reduction_callable(cls, op):
        from .aggregation import where_function

        skipna, bias = op.skipna, op.bias

        def skew(x):
            cnt = x.count()
            mean = x.mean(skipna=skipna)
            divided = (
                (x**3).mean(skipna=skipna)
                - 3 * (x**2).mean(skipna=skipna) * mean
                + 2 * mean**3
            )
            var = x.var(skipna=skipna, ddof=0)
            if isinstance(var, ENTITY_TYPE) or var > 0:
                val = where_function(var > 0, divided / var**1.5, np.nan)
            else:
                val = np.nan
            if not bias:
                val = where_function(
                    (var > 0) & (cnt > 2),
                    val * ((cnt * (cnt - 1)) ** 0.5 / (cnt - 2)),
                    np.nan,
                )
            return val

        return skew


def skew_series(
    df, axis=None, skipna=True, level=None, combine_size=None, bias=False, method=None
):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameSkew(
        axis=axis,
        skipna=skipna,
        level=level,
        combine_size=combine_size,
        bias=bias,
        output_types=[OutputType.scalar],
        use_inf_as_na=use_inf_as_na,
        method=method,
    )
    return op(df)


def skew_dataframe(
    df,
    axis=None,
    skipna=True,
    level=None,
    numeric_only=None,
    combine_size=None,
    bias=False,
    method=None,
):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameSkew(
        axis=axis,
        skipna=skipna,
        level=level,
        numeric_only=numeric_only,
        bias=bias,
        combine_size=combine_size,
        output_types=[OutputType.series],
        use_inf_as_na=use_inf_as_na,
        method=method,
    )
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...core import OutputType
from ...core.operand import OperandStage
from ..utils import gen_unknown_index_value, hash_dataframe_on
from ._duplicate import DuplicateOperand, validate_subset


class DataFrameDuplicated(DuplicateOperand):
    _op_type_ = opcodes.DUPLICATED

    def __init__(
        self,
        subset=None,
        keep=None,
        output_types=None,
        method=None,
        subset_chunk=None,
        shuffle_size=None,
        **kw
    ):
        super().__init__(
            _subset=subset,
            _keep=keep,
            _output_types=output_types,
            _method=method,
            _subset_chunk=subset_chunk,
            _shuffle_size=shuffle_size,
            **kw
        )

    @classmethod
    def _get_shape(cls, input_shape, op):
        return (input_shape[0],)

    @classmethod
    def _gen_tileable_params(cls, op: "DataFrameDuplicated", input_params):
        # duplicated() always returns a Series
        return {
            "shape": cls._get_shape(input_params["shape"], op),
            "index_value": input_params["index_value"],
            "dtype": np.dtype(bool),
            "name": input_params.get("name"),
        }

    def __call__(self, inp, inplace=False):
        self._output_types = [OutputType.series]
        params = self._gen_tileable_params(self, inp.params)

        return self.new_tileable([inp], kws=[params])

    @classmethod
    def _get_map_output_types(cls, input_chunk, method: str):
        if method in ("tree", "subset_tree"):
            return [OutputType.dataframe]
        else:
            return input_chunk.op.output_types

    @classmethod
    def _gen_chunk_params_default(cls, op: "DataFrameDuplicated", input_chunk):
        return {
            "shape": cls._get_shape(input_chunk.shape, op),
            "index_value": input_chunk.index_value,
            "dtype": np.dtype(bool),
            "name": input_chunk.name if input_chunk.ndim == 1 else None,
            "index": (input_chunk.index[0],),
        }

    @classmethod
    def _get_intermediate_shape(cls, input_shape):
        if len(input_shape) > 1:
            s = input_shape[1:]
        else:
            s = (2,)
        return (np.nan,) + s

    @classmethod
    def _gen_intermediate_chunk_params(cls, op: "DataFrameDuplicated", input_chunk):
        inp = op.input
        chunk_params = dict()
        chunk_params["shape"] = shape = cls._get_intermediate_shape(input_chunk.shape)
        chunk_params["index"] = input_chunk.index[:1] + (0,) * (len(shape) - 1)
        chunk_params["index_value"] = gen_unknown_index_value(
            input_chunk.index_value, input_chunk
        )
        if inp.ndim == 2 and len(shape) == 2:
            chunk_params["columns_value"] = input_chunk.columns_value
            chunk_params["dtypes"] = input_chunk.dtypes
        return chunk_params

    @classmethod
    def _gen_chunk_params(cls, op: "DataFrameDuplicated", input_chunk):
        is_terminal_chunk = False
        if op.method is None:
            # one chunk
            is_terminal_chunk = True
        elif op.method == "subset_tree" and op.stage is None:
            is_terminal_chunk = True
        elif op.method == "tree" and op.stage == OperandStage.agg:
            is_terminal_chunk = True
        elif op.method == "shuffle" and op.reducer_phase == "put_back":
            is_terminal_chunk = True

        if is_terminal_chunk:
            return cls._gen_chunk_params_default(op, input_chunk)
        else:
            return cls._gen_intermediate_chunk_params(op, input_chunk)

    @classmethod
    def _duplicated(cls, inp, op, subset=None, keep=None):
        if keep is None:
            keep = op.keep
        if inp.ndim == 2:
            if subset is None:
                subset = op.subset
            return inp.duplicated(subset=subset, keep=keep)
        else:
            return inp.duplicated(keep=keep)

    @classmethod
    def _execute_chunk(cls, ctx, op):
        inp = ctx[op.input.key]
        ctx[op.outputs[0].key] = cls._duplicated(inp, op)

    @classmethod
    def _execute_tree_map(cls, ctx, op):
        inp = ctx[op.input.key]
        xdf = cls._get_xdf(inp)
        if op.subset is not None:
            result = inp[op.subset].copy()
        else:
            result = inp.copy()
        duplicated = cls._duplicated(inp, op)
        if not duplicated.name:
            duplicated.name = "_duplicated_"
        result.iloc[duplicated.values] = None
        result = xdf.concat([result, duplicated], axis=1)
        ctx[op.outputs[0].key] = result

    @classmethod
    def _execute_tree_combine(cls, ctx, op):
        inp = ctx[op.input.key]
        result = inp.copy()
        duplicated_filter = ~inp.iloc[:, -1]
        duplicates = inp.loc[duplicated_filter]
        dup_on_duplicated = cls._duplicated(duplicates, op)
        result.iloc[duplicated_filter.values, -1] = dup_on_duplicated
        duplicated = result.iloc[:, -1]
        result.iloc[duplicated.values, :-1] = None
        ctx[op.outputs[0].key] = result

    @classmethod
    def _execute_tree_agg(cls, ctx, op):
        inp = ctx[op.input.key]
        result = inp.iloc[:, -1].copy()
        duplicates = inp[~inp.iloc[:, -1]]
        dup_on_duplicated = cls._duplicated(duplicates, op)
        result[~inp.iloc[:, -1]] = dup_on_duplicated
        expect_name = op.outputs[0].name
        if result.name != expect_name:
            result.name = expect_name
        result = result.astype(bool)
        ctx[op.outputs[0].key] = result

    @classmethod
    def _execute_subset_tree_post(cls, ctx, op):
        inp = ctx[op.input.key]
        idx = op.outputs[0].index[0]
        subset = ctx[op.subset_chunk.key]
        selected = subset[subset["_chunk_index_"] == idx]["_i_"]

        xdf = cls._get_xdf(inp)
        duplicated = np.ones(len(inp), dtype=bool)
        duplicated[selected] = False

        ctx[op.outputs[0].key] = xdf.Series(duplicated, index=inp.index)

    @classmethod
    def _execute_shuffle_map(cls, ctx, op):
        out = op.outputs[0]
        shuffle_size = op.shuffle_size
        subset = op.subset

        inp = ctx[op.input.key]
        if subset is not None:
            result = inp[subset].copy()
        else:
            result = inp.copy()
        if result.ndim == 1:
            name = result.name
            result = result.to_frame()
            if name is None:
                result.columns = ["_duplicated_"]
            subset = result.columns.tolist()
        else:
            if subset is None:
                subset = result.columns.tolist()
            if len(subset) == 1:
                result.columns = subset = ["_duplicated_"]
        result["_chunk_index_"] = out.index[0]
        result["_i_"] = np.arange(result.shape[0])
        hashed = hash_dataframe_on(result, subset, shuffle_size)
        for i, data in enumerate(hashed):
            reducer_idx = (i,) + out.index[1:]
            ctx[out.key, reducer_idx] = result.iloc[data]

    @classmethod
    def _execute_shuffle_reduce(cls, ctx, op: "DataFrameDuplicated"):
        out = op.outputs[0]
        inputs = list(op.iter_mapper_data(ctx))

        xdf = cls._get_xdf(inputs[0])
        inp = xdf.concat(inputs)
        subset = [c for c in inp.columns if c not in ("_chunk_index_", "_i_")]
        duplicated = cls._duplicated(inp, op, subset=subset)
        result = xdf.concat([duplicated, inp[["_chunk_index_", "_i_"]]], axis=1)
        for i in range(op.shuffle_size):
            filtered = result[result["_chunk_index_"] == i]
            del filtered["_chunk_index_"]
            if len(subset) > 1 or subset[0] == "_duplicated_":
                filtered.columns = ["_duplicated_"] + filtered.columns[1:].tolist()
            else:
                filtered.columns = [subset[0]] + filtered.columns[1:].tolist()
            ctx[out.key, (i,)] = filtered

    @classmethod
    def _execute_shuffle_put_back(cls, ctx, op: "DataFrameDuplicated"):
        inputs = list(op.iter_mapper_data(ctx))

        xdf = cls._get_xdf(inputs[0])
        inp = xdf.concat(inputs)
        inp.sort_values("_i_", inplace=True)
        del inp["_i_"]
        duplicated = inp.iloc[:, 0]
        if duplicated.name == "_duplicated_":
            duplicated.name = None
        ctx[op.outputs[0].key] = duplicated

    @classmethod
    def execute(cls, ctx, op: "DataFrameDuplicated"):
        if op.method is None:
            # one chunk
            cls._execute_chunk(ctx, op)
        elif op.method == "tree":
            # tree
            if op.stage == OperandStage.map:
                cls._execute_tree_map(ctx, op)
            elif op.stage == OperandStage.combine:
                cls._execute_tree_combine(ctx, op)
            else:
                assert op.stage == OperandStage.agg
                cls._execute_tree_agg(ctx, op)
        elif op.method == "subset_tree":
            # subset tree
            if op.stage == OperandStage.map:
                cls._execute_subset_tree_map(ctx, op)
            elif op.stage == OperandStage.combine:
                cls._execute_subset_tree_combine(ctx, op)
            elif op.stage == OperandStage.agg:
                cls._execute_subset_tree_agg(ctx, op)
            else:
                # post
                cls._execute_subset_tree_post(ctx, op)
        else:
            assert op.method == "shuffle"
            if op.stage == OperandStage.map:
                cls._execute_shuffle_map(ctx, op)
            elif op.reducer_phase == "drop_duplicates":
                cls._execute_shuffle_reduce(ctx, op)
            else:
                assert op.reducer_phase == "put_back"
                cls._execute_shuffle_put_back(ctx, op)


def df_duplicated(df, subset=None, keep="first", method="auto"):
    """
    Return boolean Series denoting duplicate rows.

    Considering certain columns is optional.

    Parameters
    ----------
    subset : column label or sequence of labels, optional
        Only consider certain columns for identifying duplicates, by
        default use all of the columns.
    keep : {'first', 'last', False}, default 'first'
        Determines which duplicates (if any) to mark.

        - ``first`` : Mark duplicates as ``True`` except for the first occurrence.
        - ``last`` : Mark duplicates as ``True`` except for the last occurrence.
        - False : Mark all duplicates as ``True``.

    Returns
    -------
    Series
        Boolean series for each duplicated rows.

    See Also
    --------
    Index.duplicated : Equivalent method on index.
    Series.duplicated : Equivalent method on Series.
    Series.drop_duplicates : Remove duplicate values from Series.
    DataFrame.drop_duplicates : Remove duplicate values from DataFrame.

    Examples
    --------
    Consider dataset containing ramen rating.

    >>> import mars.dataframe as md

    >>> df = md.DataFrame({
    ...     'brand': ['Yum Yum', 'Yum Yum', 'Indomie', 'Indomie', 'Indomie'],
    ...     'style': ['cup', 'cup', 'cup', 'pack', 'pack'],
    ...     'rating': [4, 4, 3.5, 15, 5]
    ... })
    >>> df.execute()
        brand style  rating
    0  Yum Yum   cup     4.0
    1  Yum Yum   cup     4.0
    2  Indomie   cup     3.5
    3  Indomie  pack    15.0
    4  Indomie  pack     5.0

    By default, for each set of duplicated values, the first occurrence
    is set on False and all others on True.

    >>> df.duplicated().execute()
    0    False
    1     True
    2    False
    3    False
    4    False
    dtype: bool

    By using 'last', the last occurrence of each set of duplicated values
    is set on False and all others on True.

    >>> df.duplicated(keep='last').execute()
    0     True
    1    False
    2    False
    3    False
    4    False
    dtype: bool

    By setting ``keep`` on False, all duplicates are True.

    >>> df.duplicated(keep=False).execute()
    0     True
    1     True
    2    False
    3    False
    4    False
    dtype: bool

    To find duplicates on specific column(s), use ``subset``.

    >>> df.duplicated(subset=['brand']).execute()
    0    False
    1     True
    2    False
    3     True
    4     True
    dtype: bool
    """

    if method not in ("auto", "tree", "subset_tree", "shuffle", None):
        raise ValueError(
            "method could only be one of "
            "'auto', 'tree', 'subset_tree', 'shuffle' or None"
        )
    subset = validate_subset(df, subset)
    op = DataFrameDuplicated(subset=subset, keep=keep, method=method)
    return op(df)


def series_duplicated(series, keep="first", method="auto"):
    """
    Indicate duplicate Series values.

    Duplicated values are indicated as ``True`` values in the resulting
    Series. Either all duplicates, all except the first or all except the
    last occurrence of duplicates can be indicated.

    Parameters
    ----------
    keep : {'first', 'last', False}, default 'first'
        Method to handle dropping duplicates:

        - 'first' : Mark duplicates as ``True`` except for the first
          occurrence.
        - 'last' : Mark duplicates as ``True`` except for the last
          occurrence.
        - ``False`` : Mark all duplicates as ``True``.

    Returns
    -------
    Series
        Series indicating whether each value has occurred in the
        preceding values.

    See Also
    --------
    Index.duplicated : Equivalent method on pandas.Index.
    DataFrame.duplicated : Equivalent method on pandas.DataFrame.
    Series.drop_duplicates : Remove duplicate values from Series.

    Examples
    --------
    By default, for each set of duplicated values, the first occurrence is
    set on False and all others on True:

    >>> import mars.dataframe as md

    >>> animals = md.Series(['lama', 'cow', 'lama', 'beetle', 'lama'])
    >>> animals.duplicated().execute()
    0    False
    1    False
    2     True
    3    False
    4     True
    dtype: bool

    which is equivalent to

    >>> animals.duplicated(keep='first').execute()
    0    False
    1    False
    2     True
    3    False
    4     True
    dtype: bool

    By using 'last', the last occurrence of each set of duplicated values
    is set on False and all others on True:

    >>> animals.duplicated(keep='last').execute()
    0     True
    1    False
    2     True
    3    False
    4    False
    dtype: bool

    By setting keep on ``False``, all duplicates are True:

    >>> animals.duplicated(keep=False).execute()
    0     True
    1    False
    2     True
    3    False
    4     True
    dtype: bool
    """
    if method not in ("auto", "tree", "shuffle", None):
        raise ValueError(
            "method could only be one of 'auto', 'tree', 'shuffle' or None"
        )
    op = DataFrameDuplicated(keep=keep, method=method)
    return op(series)


def index_duplicated(index, keep="first"):
    """
    Indicate duplicate index values.

    Duplicated values are indicated as ``True`` values in the resulting
    array. Either all duplicates, all except the first, or all except the
    last occurrence of duplicates can be indicated.

    Parameters
    ----------
    keep : {'first', 'last', False}, default 'first'
        The value or values in a set of duplicates to mark as missing.
        - 'first' : Mark duplicates as ``True`` except for the first
          occurrence.
        - 'last' : Mark duplicates as ``True`` except for the last
          occurrence.
        - ``False`` : Mark all duplicates as ``True``.

    Returns
    -------
    Tensor

    See Also
    --------
    Series.duplicated : Equivalent method on pandas.Series.
    DataFrame.duplicated : Equivalent method on pandas.DataFrame.
    Index.drop_duplicates : Remove duplicate values from Index.

    Examples
    --------
    By default, for each set of duplicated values, the first occurrence is
    set to False and all others to True:

    >>> import mars.dataframe as md

    >>> idx = md.Index(['lama', 'cow', 'lama', 'beetle', 'lama'])
    >>> idx.duplicated().execute()
    array([False, False,  True, False,  True])

    which is equivalent to

    >>> idx.duplicated(keep='first').execute()
    array([False, False,  True, False,  True])

    By using 'last', the last occurrence of each set of duplicated values
    is set on False and all others on True:

    >>> idx.duplicated(keep='last').execute()
    array([ True, False,  True, False, False])

    By setting keep on ``False``, all duplicates are True:

    >>> idx.duplicated(keep=False).execute()
    array([ True, False,  True, False,  True])
    """
    return index.to_series().duplicated(keep=keep).to_tensor()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import itertools
from numbers import Integral
from typing import Optional
import uuid

import numpy as np

from ...core import tile


def indexing_to_chunk_indices(output_chunk):
    """
    Compute input_indices and value_indices when read from or write to
    a tensor chunk.

    Parameters
    ----------
    output_chunk:
        A chunk in the output of the `__setitem__` op.

    Returns
    -------
        The indices in the input chunk, and value_indices in the value block
        that will be assigned.
    """
    input_indices = []  # index in the chunk of the mutable tensor
    value_indices = []  # index in the chunk of the assigned value
    for d, s in zip(output_chunk.op.indexes, output_chunk.op.inputs[0].shape):
        # expand the index (slice)
        idx = np.r_[slice(*d.indices(s)) if isinstance(d, slice) else d]
        input_indices.append(idx)
        if not isinstance(d, Integral):
            value_indices.append(np.arange(len(idx)))
    return input_indices, value_indices


def compute_output_of_indexing(tensor, tensor_index):
    """
    Compute the output information of `__{set,get}item__` on tensor for every chunk.
    """
    from ...tensor.indexing.core import process_index, calc_shape
    from ...tensor.indexing.getitem import TensorIndex

    tensor_index = process_index(tensor.ndim, tensor_index)
    output_shape = calc_shape(tensor.shape, tensor_index)

    index_tensor_op = TensorIndex(
        dtype=tensor.dtype, sparse=False, indexes=list(tensor_index)
    )
    index_tensor = tile(index_tensor_op.new_tensor([tensor], shape=tuple(output_shape)))
    output_chunks = index_tensor.chunks

    nsplits_acc = [
        np.cumsum(
            (0,)
            + tuple(
                c.shape[i]
                for c in output_chunks
                if all(idx == 0 for j, idx in enumerate(c.index) if j != i)
            )
        )
        for i in range(len(output_chunks[0].shape))
    ]
    return output_shape, output_chunks, nsplits_acc


def setitem_on_chunk_to_records(nsplits_acc, output_chunk, value, ts, is_scalar):
    """
    Turns a `__setitem__` on chunk to a list of index-value records.

    Parameters
    ----------
    nsplits_acc:
        Accumulate nsplits arrays of the output tensor chunks.

    Returns
    -------
        A list of `(index, value, timestamp)`, where `index` is the in-chunk index.
    """
    input_indices, value_indices = indexing_to_chunk_indices(output_chunk)

    # normalize assigned value
    if is_scalar:
        chunk_value = value
    else:
        chunk_value_slice = tuple(
            slice(
                nsplits_acc[i][output_chunk.index[i]],
                nsplits_acc[i][output_chunk.index[i] + 1],
            )
            for i in range(len(output_chunk.index))
        )
        chunk_value = value[chunk_value_slice]

    records = []
    for chunk_idx, value_idx in zip(
        itertools.product(*input_indices), itertools.product(*value_indices)
    ):
        new_value = chunk_value if is_scalar else chunk_value[value_idx]
        index_in_chunk = np.ravel_multi_index(
            chunk_idx, output_chunk.op.inputs[0].shape
        )
        records.append((index_in_chunk, new_value, ts))
    return records


def setitem_to_records(tensor, tensor_index, value, timestamp):
    """
    Compute the records of `__setitem__` on tensor for every chunk.

    Returns
    -------
        dict, a dict of chunk index to records in that chunk.
    """
    output_shape, output_chunks, nsplits_acc = compute_output_of_indexing(
        tensor, tensor_index
    )

    is_scalar = (
        np.isscalar(value)
        or isinstance(value, tuple)
        and tensor.dtype.fields is not None
    )
    if not is_scalar:
        value = np.broadcast_to(value, output_shape).astype(tensor.dtype)

    records = dict()
    for output_chunk in output_chunks:
        records_in_chunk = setitem_on_chunk_to_records(
            nsplits_acc, output_chunk, value, timestamp, is_scalar=is_scalar
        )
        records[output_chunk.op.inputs[0].index] = records_in_chunk
    return records


def getitem_on_chunk_to_records(nsplits_acc, output_chunk):
    """
    Turns a `__getitem__` on chunk to a list of index-value records.

    Parameters
    ----------
    nsplits_acc:
        Accumulate nsplits arrays of the output tensor chunks.

    Returns
    -------
        records: A list of `(index, value_index)`, where `index` is the in-chunk index, and
        `value_index` is the index in the final result block.
        chunk_value_shape: shape of result of this chunk.
        chunk_value_slice: index of result of this chunk in the whole result tensor.
    """
    input_indices, value_indices = indexing_to_chunk_indices(output_chunk)

    chunk_value_slice = tuple(
        slice(
            nsplits_acc[i][output_chunk.index[i]],
            nsplits_acc[i][output_chunk.index[i] + 1],
        )
        for i in range(len(output_chunk.index))
    )

    records = []
    for chunk_idx, value_idx in zip(
        itertools.product(*input_indices), itertools.product(*value_indices)
    ):
        index_in_chunk = np.ravel_multi_index(
            chunk_idx, output_chunk.op.inputs[0].shape
        )
        records.append((index_in_chunk, value_idx))
    return records, output_chunk.shape, chunk_value_slice


def getitem_to_records(tensor, tensor_index):
    """
    Compute the records of `__getitem__` on tensor for every chunk.

    Returns
    -------
        records and output_chunk dict, records is a dict of chunk index to records
        in that chunk.
    """
    output_shape, output_chunks, nsplits_acc = compute_output_of_indexing(
        tensor, tensor_index
    )

    records = dict()
    for output_chunk in output_chunks:
        records_in_chunk = getitem_on_chunk_to_records(nsplits_acc, output_chunk)
        records[output_chunk.op.inputs[0].index] = records_in_chunk
    return records, output_shape


def normalize_timestamp(timestamp=None):
    if timestamp is None:
        timestamp = np.datetime64(datetime.now())
    if isinstance(timestamp, datetime):
        timestamp = np.datetime64(timestamp)
    return timestamp


def normalize_name(name: Optional[str] = None):
    if not name:
        return str(uuid.uuid4())
    return name



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict, List, Optional


class AbstractMetaAPI(ABC):
    @abstractmethod
    async def get_chunk_meta(
        self, object_id: str, fields: List[str] = None, error: str = "raise"
    ) -> Optional[Dict]:
        """
        Get chunk meta

        Parameters
        ----------
        object_id
            Object ID
        fields
            Fields to obtain
        error
            Way to handle errors, 'raise' by default
        Returns
        -------
            Dict with fields as keys
        """



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType, recursive_tile
from ...serialization.serializables import ListField, BoolField
from ...tensor.base.sort import _validate_sort_psrs_kinds
from ...utils import calc_nsplits
from ..utils import (
    parse_index,
    validate_axis,
    build_concatenated_rows_frame,
    standardize_range_index,
)
from ..operands import DATAFRAME_TYPE
from .core import DataFrameSortOperand
from .psrs import DataFramePSRSOperandMixin, execute_sort_index


class DataFrameSortIndex(DataFrameSortOperand, DataFramePSRSOperandMixin):
    _op_type_ = OperandDef.SORT_INDEX

    level = ListField("level", default=None)
    sort_remaining = BoolField("sort_remaining", default=None)

    @classmethod
    def _tile(cls, op):
        df = op.inputs[0]

        if op.axis == 0:
            if df.chunk_shape[op.axis] == 1:
                if op.output_types[0] == OutputType.dataframe:
                    df = build_concatenated_rows_frame(df)
                    out_chunks = []
                    for chunk in df.chunks:
                        chunk_op = op.copy().reset_key()
                        out_chunks.append(
                            chunk_op.new_chunk(
                                [chunk],
                                shape=chunk.shape,
                                index=chunk.index,
                                index_value=chunk.index_value,
                                columns_value=chunk.columns_value,
                                dtypes=chunk.dtypes,
                            )
                        )
                    new_op = op.copy()
                    kws = op.outputs[0].params.copy()
                    kws["nsplits"] = df.nsplits
                    kws["chunks"] = out_chunks
                    return new_op.new_dataframes(op.inputs, **kws)
                else:
                    out_chunks = []
                    for chunk in df.chunks:
                        chunk_op = op.copy().reset_key()
                        out_chunks.append(
                            chunk_op.new_chunk(
                                [chunk],
                                shape=chunk.shape,
                                index=chunk.index,
                                index_value=chunk.index_value,
                                name=chunk.name,
                                dtype=chunk.dtype,
                            )
                        )
                    new_op = op.copy()
                    kws = op.outputs[0].params.copy()
                    kws["nsplits"] = df.nsplits
                    kws["chunks"] = out_chunks
                    return new_op.new_seriess(op.inputs, **kws)
            else:
                if op.output_types[0] == OutputType.dataframe:
                    df = build_concatenated_rows_frame(df)
                if op.na_position != "last":  # pragma: no cover
                    raise NotImplementedError("Only support puts NaNs at the end.")
                # use parallel sorting by regular sampling
                return (yield from cls._tile_psrs(op, df))
        else:
            assert op.axis == 1

            sorted_columns = list(
                df.columns_value.to_pandas().sort_values(ascending=op.ascending)
            )
            r = [(yield from recursive_tile(df[sorted_columns]))]
            if op.ignore_index:
                chunks = r[0].chunks
                yield chunks
                out = op.outputs[0]
                chunks = standardize_range_index(chunks, axis=0)
                new_op = op.copy()
                return new_op.new_dataframes(
                    op.inputs,
                    shape=out.shape,
                    chunks=chunks,
                    nsplits=calc_nsplits({c.index: c.shape for c in chunks}),
                    index_value=out.index_value,
                    columns_value=out.columns_value,
                    dtypes=out.dtypes,
                )
            return r

    @classmethod
    def execute(cls, ctx, op: "DataFrameSortIndex"):
        in_data = ctx[op.inputs[0].key]
        result = execute_sort_index(in_data, op)
        if op.nrows is not None:
            result = result.head(op.nrows)
        ctx[op.outputs[0].key] = result

    def _call_dataframe(self, df):
        if self.ignore_index:
            index_value = parse_index(pd.RangeIndex(df.shape[0]))
        else:
            index_value = df.index_value
        if self.axis == 0:
            return self.new_dataframe(
                [df],
                shape=df.shape,
                dtypes=df.dtypes,
                index_value=index_value,
                columns_value=df.columns_value,
            )
        else:
            dtypes = df.dtypes.sort_index(ascending=self.ascending)
            columns_value = parse_index(dtypes.index, store_data=True)
            return self.new_dataframe(
                [df],
                shape=df.shape,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )

    def _call_series(self, series):
        if self.axis != 0:  # pragma: no cover
            raise TypeError(f"Invalid axis: {self.axis}")
        if self.ignore_index:
            index_value = parse_index(pd.RangeIndex(series.shape[0]))
        else:
            index_value = series.index_value

        return self.new_series(
            [series],
            shape=series.shape,
            dtype=series.dtype,
            index_value=index_value,
            name=series.name,
        )

    def __call__(self, a):
        if isinstance(a, DATAFRAME_TYPE):
            self.output_types = [OutputType.dataframe]
            return self._call_dataframe(a)
        else:
            self.output_types = [OutputType.series]
            return self._call_series(a)


def sort_index(
    a,
    axis=0,
    level=None,
    ascending=True,
    inplace=False,
    kind="quicksort",
    na_position="last",
    sort_remaining=True,
    ignore_index: bool = False,
    parallel_kind="PSRS",
    psrs_kinds=None,
):
    """
    Sort object by labels (along an axis).

    Parameters
    ----------
    a : Input DataFrame or Series.
    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis along which to sort.  The value 0 identifies the rows,
        and 1 identifies the columns.
    level : int or level name or list of ints or list of level names
        If not None, sort on values in specified index level(s).
    ascending : bool, default True
        Sort ascending vs. descending.
    inplace : bool, default False
        If True, perform operation in-place.
    kind : {'quicksort', 'mergesort', 'heapsort'}, default 'quicksort'
        Choice of sorting algorithm. See also ndarray.np.sort for more
        information.  `mergesort` is the only stable algorithm. For
        DataFrames, this option is only applied when sorting on a single
        column or label.
    na_position : {'first', 'last'}, default 'last'
        Puts NaNs at the beginning if `first`; `last` puts NaNs at the end.
        Not implemented for MultiIndex.
    sort_remaining : bool, default True
        If True and sorting by level and index is multilevel, sort by other
        levels too (in order) after sorting by specified level.
    ignore_index : bool, default False
        If True, the resulting axis will be labeled 0, 1, …, n - 1.
    parallel_kind: {'PSRS'}, optional.
        Parallel sorting algorithm, for the details, refer to:
        http://csweb.cs.wfu.edu/bigiron/LittleFE-PSRS/build/html/PSRSalgorithm.html
    psrs_kinds: Sorting algorithms during PSRS algorithm.

    Returns
    -------
    sorted_obj : DataFrame or None
        DataFrame with sorted index if inplace=False, None otherwise.
    """
    if na_position not in ["last", "first"]:  # pragma: no cover
        raise TypeError(f"Invalid na_position: {na_position}")
    psrs_kinds = _validate_sort_psrs_kinds(psrs_kinds)
    axis = validate_axis(axis, a)
    level = level if isinstance(level, (list, tuple)) else [level]
    op = DataFrameSortIndex(
        level=level,
        axis=axis,
        ascending=ascending,
        inplace=inplace,
        kind=kind,
        na_position=na_position,
        sort_remaining=sort_remaining,
        ignore_index=ignore_index,
        parallel_kind=parallel_kind,
        psrs_kinds=psrs_kinds,
        gpu=a.op.is_gpu(),
    )
    sorted_a = op(a)
    if inplace:
        a.data = sorted_a.data
    else:
        return sorted_a



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Tuple, Union

import numpy as np

from ..core import MutableTensorInfo


class AbstractMutableAPI(ABC):
    @abstractmethod
    async def create_mutable_tensor(
        self,
        shape: tuple,
        dtype: Union[np.dtype, str],
        name: str = None,
        default_value: Union[int, float] = 0,
        chunk_size: Union[int, Tuple] = None,
    ) -> MutableTensorInfo:
        """
        Create a mutable tensor.

        Parameters
        ----------
        shape: tuple
            Shape of the mutable tensor.

        dtype: np.dtype or str
            Data type of the mutable tensor.

        chunk_size: int or tuple
            Chunk size of the mutable tensor.

        name: str, optional
            Name of the mutable tensor, a random name will be used if not specified.

        default_value: optional
            Default value of the mutable tensor. Default is 0.

        Returns
        -------
            MutableTensorInfo
        """

    @abstractmethod
    async def get_mutable_tensor(self, name: str) -> MutableTensorInfo:
        """
        Get the mutable tensor by name.

        Parameters
        ----------
        name: str
            Name of the mutable tensor to get.

        Returns
        -------
            MutableTensorInfo
        """

    @abstractmethod
    async def seal_mutable_tensor(self, name: str, timestamp=None):
        """
        Seal the mutable tensor by name.

        Parameters
        ----------
        name: str
            Name of the mutable tensor to seal.

        timestamp: optional
            Operations that happened before timestamp will be sealed, and later ones will be discard.

        Returns
        -------
            object
        """

    @abstractmethod
    async def read(self, name: str, index: object, timestamp=None):
        """
        Read value from mutable tensor.

        Parameters
        ----------
        name: str
            Name of mutable tensor to read.

        index:
            Index to read from the tensor.

        timestamp: optional
            Timestamp to read value that happened before then.
        """

    @abstractmethod
    async def write(self, name: str, index: object, value: object, timestamp=None):
        """
        Write value to mutable tensor.

        Parameters
        ----------
        name: str
            Name of the mutable tensor to write.

        index:
            Index to write to the tensor.

        value:
            The value that will be filled into the mutable tensor according to `index`.

        timestamp: optional
            Timestamp to associated with the newly touched value.
        """



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from ...serialization.serializables import (
    Serializable,
    FieldTypes,
    StringField,
    ListField,
    BoolField,
    Int32Field,
)
from ...typing import BandType


class SubtaskScheduleSummary(Serializable):
    task_id: str = StringField("task_id")
    subtask_id: str = StringField("subtask_id")
    bands: List[BandType] = ListField("bands", FieldTypes.tuple(FieldTypes.string))
    is_finished: bool = BoolField("is_finished", default=False)
    is_cancelled: bool = BoolField("is_cancelled", default=False)
    num_reschedules: int = Int32Field("num_reschedules", default=0)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ...core import NodeRole, AbstractService
from ..procinfo import ProcessInfoManagerActor
from ..uploader import NodeInfoUploaderActor
from .locator import SupervisorPeerLocatorActor
from .node_allocator import NodeAllocatorActor
from .node_info import NodeInfoCollectorActor


class ClusterSupervisorService(AbstractService):
    """
    Cluster service on supervisor

    Service Configuration
    ---------------------
    {
        "cluster": {
            "backend": "<cluster backend name>",
            "lookup_address": "<address of master>",
            "node_timeout": timeout seconds of nodes,
            "node_check_interval": check interval seconds for nodes
        }
    }
    """

    async def start(self):
        svc_config = self._config["cluster"]
        address = self._address

        backend = svc_config.get("backend", "fixed")
        lookup_address = svc_config.get(
            "lookup_address", address if backend == "fixed" else None
        )
        await mo.create_actor(
            NodeInfoCollectorActor,
            timeout=svc_config.get("node_timeout"),
            check_interval=svc_config.get("node_check_interval"),
            uid=NodeInfoCollectorActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            SupervisorPeerLocatorActor,
            backend_name=backend,
            lookup_address=lookup_address,
            uid=SupervisorPeerLocatorActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            NodeInfoUploaderActor,
            role=NodeRole.SUPERVISOR,
            interval=svc_config.get("node_check_interval"),
            uid=NodeInfoUploaderActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            NodeAllocatorActor,
            backend_name=backend,
            lookup_address=lookup_address,
            uid=NodeAllocatorActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            ProcessInfoManagerActor,
            uid=ProcessInfoManagerActor.default_uid(),
            address=address,
        )

    async def stop(self):
        address = self._address

        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=NodeInfoCollectorActor.default_uid(), address=address
            )
        )
        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=SupervisorPeerLocatorActor.default_uid(), address=address
            )
        )
        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=NodeInfoUploaderActor.default_uid(), address=address
            )
        )
        await mo.destroy_actor(
            mo.create_actor_ref(uid=NodeAllocatorActor.default_uid(), address=address)
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import functools
from typing import Optional

import numpy as np
import pandas as pd

from ... import opcodes, options, get_context
from ...core import recursive_tile, OutputType
from ...core.operand import OperandStage
from ...serialization.serializables import Int64Field, ReferenceField
from ...utils import ceildiv, has_unknown_shape
from ..datasource import tensor as astensor
from ..operands import TensorOperandMixin, TensorMapReduceOperand


class TensorBinCount(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = opcodes.BINCOUNT

    weights = ReferenceField("weights", default=None)
    minlength: Optional[int] = Int64Field("minlength", default=0)
    chunk_size_limit: int = Int64Field("chunk_size_limit")

    chunk_count: Optional[int] = Int64Field("chunk_count")
    tileable_right_bound: Optional[int] = Int64Field("tileable_right_bound")

    def __call__(self, x, weights=None):
        inputs = [x]
        self.weights = weights
        dtype = np.dtype(np.int_)
        if weights is not None:
            inputs.append(weights)
            dtype = weights.dtype
        return self.new_tensor(inputs, dtype=dtype, shape=(np.nan,))

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if len(inputs) > 1:
            self.weights = inputs[1]

    @classmethod
    def _tile_single(cls, op: "TensorBinCount"):
        out = op.outputs[0]
        new_chunk_op = op.copy().reset_key()
        chunk_inputs = [op.inputs[0].chunks[0]]
        if op.weights is not None:
            chunk_inputs.append(op.weights.chunks[0])
        new_chunk = new_chunk_op.new_chunk(chunk_inputs, index=(0,), **out.params)

        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            op.inputs, chunks=[new_chunk], nsplits=((np.nan,),), **out.params
        )

    @classmethod
    def tile(cls, op: "TensorBinCount"):
        from ...dataframe.operands import DataFrameShuffleProxy
        from ...dataframe.utils import parse_index

        if has_unknown_shape(*op.inputs):
            yield

        ctx = get_context()
        a = op.inputs[0]
        out = op.outputs[0]

        if op.weights is not None and a.shape != op.weights.shape:
            raise ValueError("The weights and list don't have the same length.")

        input_max = yield from recursive_tile(a.max())
        yield input_max.chunks + [c for inp in op.inputs for c in inp.chunks]
        [max_val] = ctx.get_chunks_result([input_max.chunks[0].key])
        tileable_right_bound = max(op.minlength, int(max_val) + 1)

        chunk_count = max(1, ceildiv(tileable_right_bound, op.chunk_size_limit))

        if (
            len(op.inputs[0].chunks) == 1
            and (op.weights is None or len(op.weights.chunks) == 1)
            and chunk_count == 1
        ):
            return cls._tile_single(op)

        if op.weights is not None:
            weights = yield from recursive_tile(op.weights.rechunk(a.nsplits))
            weights_chunks = weights.chunks
        else:
            weights_chunks = itertools.repeat(None)

        map_chunks = []
        for a_chunk, weights_chunk in zip(a.chunks, weights_chunks):
            new_op = op.copy().reset_key()
            new_op.chunk_count = chunk_count
            new_op.tileable_right_bound = tileable_right_bound
            new_op.stage = OperandStage.map
            new_op._output_types = [OutputType.series]

            inputs = [a_chunk]
            if weights_chunk is not None:
                inputs.append(weights_chunk)
            map_chunks.append(
                new_op.new_chunk(
                    inputs,
                    dtype=out.dtype,
                    shape=(np.nan,),
                    index=a_chunk.index,
                    index_value=parse_index(pd.Index([0], dtype=np.int64), a_chunk.key),
                )
            )

        shuffle_op = DataFrameShuffleProxy(output_types=[OutputType.tensor]).new_chunk(
            map_chunks, dtype=out.dtype, shape=()
        )

        reduce_chunks = []
        reduce_nsplits = []
        left_offset = 0
        for chunk_idx in range(chunk_count):
            right_offset = min(tileable_right_bound, left_offset + op.chunk_size_limit)

            new_op = op.copy().reset_key()
            new_op.stage = OperandStage.reduce
            new_op.reducer_ordinal = chunk_idx
            new_op.n_reducers = chunk_count
            new_op.chunk_count = chunk_count
            new_op.tileable_right_bound = tileable_right_bound

            reduce_chunks.append(
                new_op.new_chunk(
                    [shuffle_op],
                    dtype=out.dtype,
                    shape=(right_offset - left_offset,),
                    index=(chunk_idx,),
                )
            )
            reduce_nsplits.append(right_offset - left_offset)
            left_offset = right_offset

        new_op = op.copy().reset_key()
        params = out.params.copy()
        params["shape"] = (tileable_right_bound,)
        return new_op.new_tileables(
            op.inputs,
            chunks=reduce_chunks,
            nsplits=(tuple(reduce_nsplits),),
            **params,
        )

    @classmethod
    def _execute_map(cls, ctx, op: "TensorBinCount"):
        input_val = ctx[op.inputs[0].key]
        if op.weights is not None:
            weights_val = ctx[op.weights.key]
            df = pd.DataFrame({"data": input_val, "weights": weights_val})
            res = df.groupby("data")["weights"].sum()
        else:
            res = pd.Series(input_val).groupby(input_val).count()

        if res.index.min() < 0:
            raise ValueError("'list' argument must have no negative elements")

        left_bound = 0
        for target_idx in range(op.chunk_count):
            right_bound = res.index.searchsorted(
                (1 + target_idx) * op.chunk_size_limit, "left"
            )
            sliced = res.iloc[left_bound:right_bound]
            if len(sliced) > 0:
                ctx[op.outputs[0].key, (target_idx,)] = sliced
            left_bound = right_bound

    @classmethod
    def _execute_reduce(cls, ctx, op: "TensorBinCount"):
        out = op.outputs[0]
        input_list = []
        for input_key in op.inputs[0].op.source_keys:
            sliced = ctx.get((input_key, out.index))
            if sliced is not None:
                input_list.append(sliced)

        left_bound = op.chunk_size_limit * out.index[0]
        right_bound = min(left_bound + op.chunk_size_limit, op.tileable_right_bound)
        if not input_list:
            ctx[op.outputs[0].key] = np.zeros(right_bound - left_bound)
        else:
            res = functools.reduce(
                lambda a, b: a.add(b, fill_value=0), input_list
            ).astype(out.dtype)
            res = res.reindex(pd.RangeIndex(left_bound, right_bound), fill_value=0)
            ctx[op.outputs[0].key] = res.values

    @classmethod
    def execute(cls, ctx, op: "TensorBinCount"):
        if op.stage == OperandStage.map:
            op._execute_map(ctx, op)
        elif op.stage == OperandStage.reduce:
            op._execute_reduce(ctx, op)
        else:
            input_val = ctx[op.inputs[0].key]
            weights_val = ctx[op.weights.key] if op.weights is not None else None
            ctx[op.outputs[0].key] = np.bincount(
                input_val, weights=weights_val, minlength=op.minlength
            )


def bincount(x, weights=None, minlength=0, chunk_size_limit=None):
    """
    Count number of occurrences of each value in array of non-negative ints.

    The number of bins (of size 1) is one larger than the largest value in
    `x`. If `minlength` is specified, there will be at least this number
    of bins in the output array (though it will be longer if necessary,
    depending on the contents of `x`).
    Each bin gives the number of occurrences of its index value in `x`.
    If `weights` is specified the input array is weighted by it, i.e. if a
    value ``n`` is found at position ``i``, ``out[n] += weight[i]`` instead
    of ``out[n] += 1``.

    Parameters
    ----------
    x : tensor or array_like, 1 dimension, nonnegative ints
        Input array.
    weights : tensor or array_like, optional
        Weights, array of the same shape as `x`.
    minlength : int, optional
        A minimum number of bins for the output array.

    Returns
    -------
    out : tensor of ints
        The result of binning the input array.
        The length of `out` is equal to ``np.amax(x)+1``.

    Raises
    ------
    ValueError
        If the input is not 1-dimensional, or contains elements with negative
        values, or if `minlength` is negative.
    TypeError
        If the type of the input is float or complex.

    See Also
    --------
    histogram, digitize, unique

    Examples
    --------
    >>> import mars.tensor as mt
    >>> mt.bincount(mt.arange(5)).execute()
    array([1, 1, 1, 1, 1])
    >>> mt.bincount(mt.tensor([0, 1, 1, 3, 2, 1, 7])).execute()
    array([1, 3, 1, 1, 0, 0, 0, 1])

    The input array needs to be of integer dtype, otherwise a
    TypeError is raised:

    >>> mt.bincount(mt.arange(5, dtype=float)).execute()
    Traceback (most recent call last):
      ....execute()
    TypeError: Cannot cast array data from dtype('float64') to dtype('int64')
    according to the rule 'safe'

    A possible use of ``bincount`` is to perform sums over
    variable-size chunks of an array, using the ``weights`` keyword.

    >>> w = mt.array([0.3, 0.5, 0.2, 0.7, 1., -0.6]) # weights
    >>> x = mt.array([0, 1, 1, 2, 2, 2])
    >>> mt.bincount(x,  weights=w).execute()
    array([ 0.3,  0.7,  1.1])
    """
    x = astensor(x)
    weights = astensor(weights) if weights is not None else None

    if not np.issubdtype(x.dtype, np.int_):
        raise TypeError(f"Cannot cast array data from {x.dtype} to {np.dtype(np.int_)}")
    if x.ndim != 1:
        raise ValueError("'x' must be 1 dimension")
    if minlength < 0:
        raise ValueError("'minlength' must not be negative")

    chunk_size_limit = (
        chunk_size_limit
        if chunk_size_limit is not None
        else options.bincount.chunk_size_limit
    )
    op = TensorBinCount(minlength=minlength, chunk_size_limit=chunk_size_limit)
    return op(x, weights=weights)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
from typing import Dict

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import (
    FieldTypes,
    KeyField,
    StringField,
    BytesField,
    TupleField,
)
from ...lib.filesystem import get_fs, FSMap
from ...utils import has_unknown_shape
from .core import TensorDataStore


class ZarrOptions(object):
    def __init__(self, options: Dict):
        self._options = options

    def todict(self):
        return self._options

    @staticmethod
    def _stringfy(v):
        return pickle.dumps(v) if not isinstance(v, str) else v

    def __mars_tokenize__(self):
        return (
            list(self._options.keys()),
            list(self._stringfy(v) for v in self._options.values()),
        )

    def __getstate__(self):
        return self._options

    def __setstate__(self, state):
        self._options = state


class TensorToZarrDataStore(TensorDataStore):
    _op_type_ = OperandDef.TENSOR_STORE_ZARR

    _input = KeyField("input")
    _path = StringField("path")
    _group = StringField("group")
    _dataset = StringField("dataset")
    _zarr_options = BytesField(
        "zarr_options", on_serialize=pickle.dumps, on_deserialize=pickle.loads
    )
    _axis_offsets = TupleField("axis_offsets", FieldTypes.int32)

    def __init__(
        self,
        path=None,
        group=None,
        dataset=None,
        zarr_options=None,
        axis_offsets=None,
        **kw,
    ):
        super().__init__(
            _path=path,
            _group=group,
            _dataset=dataset,
            _zarr_options=zarr_options,
            _axis_offsets=axis_offsets,
            **kw,
        )

    @property
    def path(self):
        return self._path

    @property
    def group(self):
        return self._group

    @property
    def dataset(self):
        return self._dataset

    @property
    def zarr_options(self):
        return self._zarr_options

    @property
    def axis_offsets(self):
        return self._axis_offsets

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        import zarr

        if has_unknown_shape(*op.inputs):
            yield
        in_tensor = op.input

        # create dataset
        fs = get_fs(op.path, None)
        path = op.path
        if op.group is not None:
            path += "/" + op.group
        fs_map = FSMap(path, fs)
        zarr.open(
            fs_map,
            "w",
            path=op.dataset,
            dtype=in_tensor.dtype,
            shape=in_tensor.shape,
            chunks=tuple(max(ns) for ns in in_tensor.nsplits),
            **op.zarr_options.todict(),
        )

        cum_nsplits = [[0] + np.cumsum(ns).tolist() for ns in in_tensor.nsplits]
        out_chunks = []
        for chunk in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._axis_offsets = tuple(
                cs[i] for i, cs in zip(chunk.index, cum_nsplits)
            )
            out_chunks.append(
                chunk_op.new_chunk([chunk], shape=(0,) * chunk.ndim, index=chunk.index)
            )

        new_op = op.copy()
        out = op.outputs[0]
        nsplits = tuple((0,) * len(ns) for ns in in_tensor.nsplits)
        return new_op.new_tensors(
            op.inputs,
            shape=out.shape,
            order=out.order,
            nsplits=nsplits,
            chunks=out_chunks,
        )

    @classmethod
    def execute(cls, ctx, op):
        import zarr

        fs = get_fs(op.path, None)
        fs_map = FSMap(op.path, fs)

        group = zarr.Group(store=fs_map, path=op.group)
        array = group[op.dataset]

        to_store = ctx[op.inputs[0].key]
        axis_offsets = op.axis_offsets
        shape = to_store.shape

        array[
            tuple(
                slice(offset, offset + size)
                for offset, size in zip(axis_offsets, shape)
            )
        ] = to_store

        ctx[op.outputs[0].key] = np.empty((0,) * to_store.ndim, dtype=to_store.dtype)


def tozarr(path, x, group=None, dataset=None, **zarr_options):
    import zarr

    if isinstance(path, zarr.Array):
        arr = path
        if isinstance(arr.store, FSMap):
            root = arr.store.root
            path, dataset = root.rsplit("/", 1)
        else:
            path = arr.store.path
            if "/" in arr.path and group is None:
                group = arr.path.rsplit("/", 1)[0]
            dataset = arr.basename
            if not dataset:
                path, dataset = path.rsplit("/", 1)
        for attr in ["compressor", "filters"]:
            if getattr(arr, attr):
                zarr_options[attr] = getattr(arr, attr)
    elif isinstance(path, str):
        if dataset is None:
            path, dataset = path.rsplit("/", 1)
    else:
        raise TypeError(
            "`path` passed has wrong type, "
            "expect str, or zarr.Array"
            f"got {type(path)}"
        )

    op = TensorToZarrDataStore(
        path=path, group=group, dataset=dataset, zarr_options=ZarrOptions(zarr_options)
    )
    return op(x)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    import tiledb
except (ImportError, OSError):  # pragma: no cover
    tiledb = None


def get_tiledb_schema_from_tensor(tensor, tiledb_ctx, nsplits, **kw):
    from ..core import TensorOrder

    ctx = tiledb_ctx

    dims = []
    for d in range(tensor.ndim):
        extent = tensor.shape[d]
        domain = (0, extent - 1)
        tile = max(nsplits[d])
        dims.append(
            tiledb.Dim(name="", domain=domain, tile=tile, dtype=np.int64, ctx=ctx)
        )
    dom = tiledb.Domain(*dims, ctx=ctx)
    att = tiledb.Attr(ctx=ctx, dtype=tensor.dtype)
    cell_order = "C" if tensor.order == TensorOrder.C_ORDER else "F"
    return tiledb.ArraySchema(
        ctx=ctx,
        domain=dom,
        attrs=(att,),
        sparse=tensor.issparse(),
        cell_order=cell_order,
        **kw,
    )


def check_tiledb_array_with_tensor(tensor, tiledb_array):
    if tensor.ndim != tiledb_array.ndim:
        # ndim
        raise ValueError(
            "ndim of TileDB Array to store is different to tensor, "
            f"expect {tensor.ndim}, got {tiledb_array.ndim}"
        )
    if tensor.shape != tiledb_array.shape:
        # shape
        raise ValueError(
            "shape of TileDB Array to store is different to tensor, "
            f"expect {tensor.shape}, got {tiledb_array.shape}"
        )
    if tensor.dtype != tiledb_array.attr(0).dtype:
        # dtype
        raise ValueError(
            "dtype of TileDB Array to store is different to tensor, "
            f"expect {tensor.dtype}, got {tiledb_array.domain.dtype}"
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .chunks import Chunk, ChunkData, CHUNK_TYPE
from .core import Entity, EntityData, ENTITY_TYPE
from .executable import ExecutableTuple, _ExecuteAndFetchMixin
from .fuse import FuseChunk, FuseChunkData, FUSE_CHUNK_TYPE
from .objects import (
    ObjectChunk,
    ObjectChunkData,
    Object,
    ObjectData,
    OBJECT_CHUNK_TYPE,
    OBJECT_TYPE,
)
from .output_types import (
    OutputType,
    register_output_types,
    get_output_types,
    register_fetch_class,
    get_fetch_class,
    get_tileable_types,
    get_chunk_types,
)
from .tileables import (
    Tileable,
    TileableData,
    TILEABLE_TYPE,
    HasShapeTileable,
    HasShapeTileableData,
    NotSupportTile,
    register,
    unregister,
)
from .utils import tile, recursive_tile



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import operator
import pprint
import sys
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional

from .... import oscar as mo
from ....core import ExecutionError
from ....core.graph import DAG
from ....core.operand import Fetch, FetchShuffle
from ....lib.aio import alru_cache
from ....metrics import Metrics
from ....oscar.errors import MarsError
from ....storage import StorageLevel
from ....utils import dataslots, get_chunk_key_to_data_keys, wrap_exception
from ...cluster import ClusterAPI
from ...meta import MetaAPI
from ...storage import StorageAPI
from ...subtask import Subtask, SubtaskAPI, SubtaskResult, SubtaskStatus
from .workerslot import BandSlotManagerActor
from .quota import QuotaActor

logger = logging.getLogger(__name__)

# the default times to run subtask.
DEFAULT_SUBTASK_MAX_RETRIES = 0


@dataslots
@dataclass
class SubtaskExecutionInfo:
    aio_task: asyncio.Task
    band_name: str
    supervisor_address: str
    result: SubtaskResult = field(default_factory=SubtaskResult)
    cancelling: bool = False
    max_retries: int = 0
    num_retries: int = 0
    slot_id: Optional[int] = None
    kill_timeout: Optional[int] = None


async def _retry_run(
    subtask: Subtask, subtask_info: SubtaskExecutionInfo, target_async_func, *args
):
    assert subtask_info.num_retries >= 0
    assert subtask_info.max_retries >= 0

    while True:
        try:
            return await target_async_func(*args)
        except (OSError, MarsError) as ex:
            if subtask_info.num_retries < subtask_info.max_retries:
                logger.error(
                    "Rerun[%s/%s] the %s of subtask %s due to %s.",
                    subtask_info.num_retries,
                    subtask_info.max_retries,
                    target_async_func,
                    subtask.subtask_id,
                    ex,
                )
                subtask_info.num_retries += 1
                continue
            if subtask_info.max_retries > 0:
                message = (
                    f"Exceed max rerun[{subtask_info.num_retries}/{subtask_info.max_retries}]:"
                    f" {target_async_func} of subtask {subtask.subtask_id} due to {ex}."
                )
                logger.error(message)

                raise wrap_exception(ex, wrap_name="_ExceedMaxRerun", message=message)
            else:
                raise ex
        except asyncio.CancelledError:
            raise
        except Exception as ex:
            if subtask_info.max_retries > 0:
                message = (
                    f"Failed to rerun the {target_async_func} of subtask {subtask.subtask_id}, "
                    f"num_retries: {subtask_info.num_retries}, max_retries: {subtask_info.max_retries} "
                    f"due to unhandled exception: {ex}."
                )
                logger.error(message)

                raise wrap_exception(
                    ex, wrap_name="_UnhandledException", message=message
                )
            else:
                raise ex


def _fill_subtask_result_with_exception(
    subtask: Subtask, subtask_info: SubtaskExecutionInfo
):
    _, exc, tb = sys.exc_info()
    if isinstance(exc, ExecutionError):
        exc = exc.nested_error
        tb = exc.__traceback__

    exc_info = (type(exc), exc, tb)
    if isinstance(exc, asyncio.CancelledError):
        status = SubtaskStatus.cancelled
        logger.exception(
            "Cancel run subtask %s on band %s",
            subtask.subtask_id,
            subtask_info.band_name,
            exc_info=exc_info,
        )
    else:
        status = SubtaskStatus.errored
        logger.exception(
            "Failed to run subtask %s on band %s",
            subtask.subtask_id,
            subtask_info.band_name,
            exc_info=exc_info,
        )
    subtask_info.result.status = status
    subtask_info.result.progress = 1.0
    subtask_info.result.error = exc
    subtask_info.result.traceback = tb


class SubtaskExecutionActor(mo.StatelessActor):
    _subtask_info: Dict[str, SubtaskExecutionInfo]

    def __init__(
        self,
        subtask_max_retries: int = DEFAULT_SUBTASK_MAX_RETRIES,
        enable_kill_slot: bool = True,
        data_prepare_timeout: int = 600,
    ):
        self._cluster_api = None
        self._global_resource_ref = None
        self._subtask_max_retries = subtask_max_retries
        self._enable_kill_slot = enable_kill_slot
        self._data_prepare_timeout = data_prepare_timeout

        self._subtask_info = dict()
        self._submitted_subtask_count = Metrics.counter(
            "mars.band.submitted_subtask_count",
            "The count of submitted subtasks to the current band.",
            ("band",),
        )
        self._finished_subtask_count = Metrics.counter(
            "mars.band.finished_subtask_count",
            "The count of finished subtasks of the current band.",
            ("band",),
        )

    async def __post_create__(self):
        self._cluster_api = await ClusterAPI.create(self.address)

    @alru_cache(cache_exceptions=False)
    async def _get_slot_manager_ref(
        self, band: str
    ) -> mo.ActorRefType[BandSlotManagerActor]:
        return await mo.actor_ref(
            BandSlotManagerActor.gen_uid(band), address=self.address
        )

    @alru_cache(cache_exceptions=False)
    async def _get_band_quota_ref(self, band: str) -> mo.ActorRefType[QuotaActor]:
        return await mo.actor_ref(QuotaActor.gen_uid(band), address=self.address)

    async def _prepare_input_data(self, subtask: Subtask, band_name: str):
        queries = []
        shuffle_queries = []
        storage_api = await StorageAPI.create(
            subtask.session_id, address=self.address, band_name=band_name
        )
        chunk_key_to_data_keys = get_chunk_key_to_data_keys(subtask.chunk_graph)
        for chunk in subtask.chunk_graph:
            if chunk.key in subtask.pure_depend_keys:
                continue
            if chunk.op.gpu:  # pragma: no cover
                to_fetch_band = band_name
            else:
                to_fetch_band = "numa-0"
            if isinstance(chunk.op, Fetch):
                queries.append(
                    storage_api.fetch.delay(chunk.key, band_name=to_fetch_band)
                )
            elif isinstance(chunk.op, FetchShuffle):
                for key in chunk_key_to_data_keys[chunk.key]:
                    shuffle_queries.append(
                        storage_api.fetch.delay(
                            key, band_name=to_fetch_band, error="ignore"
                        )
                    )
        if queries:
            await storage_api.fetch.batch(*queries)
        if shuffle_queries:
            # TODO(hks): The batch method doesn't accept different error arguments,
            #  combine them when it can.

            # shuffle data fetch from remote won't be recorded in meta,
            # thus they are not tracked by lifecycle service,
            # here return remote mapper keys to remove them later.
            return await storage_api.fetch.batch(*shuffle_queries)

    async def _collect_input_sizes(
        self, subtask: Subtask, supervisor_address: str, band_name: str
    ):
        graph = subtask.chunk_graph
        sizes = dict()

        fetch_keys = list(
            set(
                n.key
                for n in graph.iter_indep()
                if isinstance(n.op, Fetch) and n.key not in subtask.pure_depend_keys
            )
        )
        if not fetch_keys:
            return sizes

        storage_api = await StorageAPI.create(
            subtask.session_id, address=self.address, band_name=band_name
        )
        meta_api = await MetaAPI.create(subtask.session_id, address=supervisor_address)

        fetch_metas = await meta_api.get_chunk_meta.batch(
            *(
                meta_api.get_chunk_meta.delay(k, fields=["memory_size", "store_size"])
                for k in fetch_keys
            )
        )
        data_infos = await storage_api.get_infos.batch(
            *(storage_api.get_infos.delay(k) for k in fetch_keys)
        )

        # compute memory quota size. when data located in shared memory, the cost
        # should be differences between deserialized memory cost and serialized cost,
        # otherwise we should take deserialized memory cost
        for key, meta, infos in zip(fetch_keys, fetch_metas, data_infos):
            level = functools.reduce(operator.or_, (info.level for info in infos))
            if level & StorageLevel.MEMORY:
                mem_cost = max(0, meta["memory_size"] - meta["store_size"])
            else:
                mem_cost = meta["memory_size"]
            sizes[key] = (meta["store_size"], mem_cost)

        return sizes

    @classmethod
    def _estimate_sizes(cls, subtask: Subtask, input_sizes: Dict):
        size_context = dict(input_sizes.items())
        graph = subtask.chunk_graph

        key_to_ops = defaultdict(set)
        chunk_key_to_sizes = defaultdict(lambda: 0)
        for n in graph:
            key_to_ops[n.op.key].add(n.op)
            chunk_key_to_sizes[n.key] += 1
        key_to_ops = {k: list(v) for k, v in key_to_ops.items()}

        # condense op key graph
        op_key_graph = DAG()
        for n in graph.topological_iter():
            if n.key in subtask.pure_depend_keys:
                continue
            if n.op.key not in op_key_graph:
                op_key_graph.add_node(n.op.key)
            for succ in graph.iter_successors(n):
                if succ.op.key not in op_key_graph:
                    op_key_graph.add_node(succ.op.key)
                op_key_graph.add_edge(n.op.key, succ.op.key)

        key_stack = list(op_key_graph.iter_indep())
        pred_ref_count = {k: op_key_graph.count_predecessors(k) for k in op_key_graph}
        succ_ref_count = {k: op_key_graph.count_successors(k) for k in op_key_graph}

        visited_op_keys = set()
        total_memory_cost = 0
        max_memory_cost = sum(calc_size for _, calc_size in size_context.values())
        while key_stack:
            key = key_stack.pop()
            op = key_to_ops[key][0]

            if not isinstance(op, Fetch):
                op.estimate_size(size_context, op)

            calc_cost = sum(size_context[out.key][1] for out in op.outputs)
            total_memory_cost += calc_cost
            max_memory_cost = max(total_memory_cost, max_memory_cost)

            if not isinstance(op, Fetch):
                # when calculation result is stored, memory cost of calculation
                #  can be replaced with result memory cost
                result_cost = sum(size_context[out.key][0] for out in op.outputs)
                total_memory_cost += result_cost - calc_cost

            visited_op_keys.add(key)

            for succ_op_key in op_key_graph.iter_successors(key):
                pred_ref_count[succ_op_key] -= 1
                if pred_ref_count[succ_op_key] == 0:
                    key_stack.append(succ_op_key)

            for pred_op_key in op_key_graph.iter_predecessors(key):
                succ_ref_count[pred_op_key] -= 1
                if succ_ref_count[pred_op_key] == 0:
                    pred_op = key_to_ops[pred_op_key][0]
                    outs = key_to_ops[pred_op_key][0].outputs
                    for out in outs:
                        chunk_key_to_sizes[out.key] -= 1
                    # when clearing fetches, subtract memory size, otherwise subtract store size
                    account_idx = 1 if isinstance(pred_op, Fetch) else 0
                    pop_result_cost = 0
                    for out in outs:
                        # corner case exist when a fetch op and another op has same chunk key
                        # but their op keys are different
                        if chunk_key_to_sizes[out.key] == 0:
                            pop_result_cost += size_context.pop(out.key, (0, 0))[
                                account_idx
                            ]
                        else:
                            pop_result_cost += size_context.get(out.key, (0, 0))[
                                account_idx
                            ]
                    total_memory_cost -= pop_result_cost
        return sum(t[0] for t in size_context.values()), max_memory_cost

    @classmethod
    def _check_cancelling(cls, subtask_info: SubtaskExecutionInfo):
        if subtask_info.cancelling:
            raise asyncio.CancelledError

    async def _remove_mapper_data(
        self, session_id: str, band_name: str, remote_mapper_keys: List
    ):
        storage_api = await StorageAPI.create(
            session_id, address=self.address, band_name=band_name
        )
        logger.debug("Delete mapper data %s", remote_mapper_keys)
        await storage_api.delete.batch(
            *[
                storage_api.delete.delay(key, error="ignore")
                for key in remote_mapper_keys
            ]
        )

    async def internal_run_subtask(self, subtask: Subtask, band_name: str):
        subtask_api = SubtaskAPI(self.address)
        subtask_info = self._subtask_info[subtask.subtask_id]
        subtask_info.result = SubtaskResult(
            subtask_id=subtask.subtask_id,
            session_id=subtask.session_id,
            task_id=subtask.task_id,
            stage_id=subtask.stage_id,
            status=SubtaskStatus.pending,
        )
        try:
            logger.debug("Preparing data for subtask %s", subtask.subtask_id)
            prepare_data_task = asyncio.create_task(
                _retry_run(
                    subtask, subtask_info, self._prepare_input_data, subtask, band_name
                )
            )
            remote_mapper_keys = await asyncio.wait_for(
                prepare_data_task, timeout=self._data_prepare_timeout
            )

            input_sizes = await self._collect_input_sizes(
                subtask, subtask_info.supervisor_address, band_name
            )
            _store_size, calc_size = await asyncio.to_thread(
                self._estimate_sizes, subtask, input_sizes
            )
            self._check_cancelling(subtask_info)

            batch_quota_req = {(subtask.session_id, subtask.subtask_id): calc_size}
            logger.debug("Start actual running of subtask %s", subtask.subtask_id)
            subtask_info.result = await self._retry_run_subtask(
                subtask, band_name, subtask_api, batch_quota_req
            )
            if remote_mapper_keys:
                await self.ref()._remove_mapper_data.tell(
                    subtask.session_id, band_name, remote_mapper_keys
                )
        except:  # noqa: E722  # pylint: disable=bare-except
            _fill_subtask_result_with_exception(subtask, subtask_info)
        finally:
            # make sure new slot usages are uploaded in time
            try:
                slot_manager_ref = await self._get_slot_manager_ref(band_name)
                await slot_manager_ref.upload_slot_usages(periodical=False)
            except:  # noqa: E722  # pylint: disable=bare-except
                _fill_subtask_result_with_exception(subtask, subtask_info)
            finally:
                # pop the subtask info at the end is to cancel the job.
                self._subtask_info.pop(subtask.subtask_id, None)
        return subtask_info.result

    async def _retry_run_subtask(
        self, subtask: Subtask, band_name: str, subtask_api: SubtaskAPI, batch_quota_req
    ):
        quota_ref = await self._get_band_quota_ref(band_name)
        slot_manager_ref = await self._get_slot_manager_ref(band_name)
        subtask_info = self._subtask_info[subtask.subtask_id]
        assert subtask_info.num_retries >= 0
        assert subtask_info.max_retries >= 0

        async def _run_subtask_once():
            aiotask = None
            slot_id = None
            try:
                await quota_ref.request_batch_quota(batch_quota_req)
                self._check_cancelling(subtask_info)

                slot_id = await slot_manager_ref.acquire_free_slot(
                    (subtask.session_id, subtask.subtask_id)
                )
                subtask_info.slot_id = slot_id
                self._check_cancelling(subtask_info)

                subtask_info.result.status = SubtaskStatus.running
                aiotask = asyncio.create_task(
                    subtask_api.run_subtask_in_slot(band_name, slot_id, subtask)
                )
                return await asyncio.shield(aiotask)
            except asyncio.CancelledError as ex:
                try:
                    if aiotask is not None:
                        logger.info(
                            "Start to cancel subtask %s in slot %s on band %s.",
                            subtask.subtask_id,
                            slot_id,
                            band_name,
                        )
                        await asyncio.wait_for(
                            asyncio.shield(
                                subtask_api.cancel_subtask_in_slot(band_name, slot_id)
                            ),
                            subtask_info.kill_timeout,
                        )
                except asyncio.TimeoutError:
                    logger.info(
                        "Wait for subtask to cancel timed out (%s). "
                        "Start killing slot %d",
                        subtask_info.kill_timeout,
                        slot_id,
                    )
                    await slot_manager_ref.kill_slot(slot_id)
                    sub_pool_address = await slot_manager_ref.get_slot_address(slot_id)
                    await mo.wait_actor_pool_recovered(sub_pool_address, self.address)
                except:  # pragma: no cover
                    logger.exception("Unexpected errors raised when handling cancel")
                    raise
                finally:
                    raise ex
            except (OSError, MarsError) as ex:
                if slot_id is not None:
                    # may encounter subprocess memory error
                    sub_pool_address = await slot_manager_ref.get_slot_address(slot_id)
                    await mo.wait_actor_pool_recovered(sub_pool_address, self.address)
                raise ex
            finally:
                # make sure allocated slots are traced
                if slot_id is None:  # pragma: no cover
                    slot_id = await slot_manager_ref.get_subtask_slot(
                        (subtask.session_id, subtask.subtask_id)
                    )
                logger.debug(
                    "Subtask %s running ended, slot_id=%r", subtask.subtask_id, slot_id
                )
                if slot_id is not None:
                    await slot_manager_ref.release_free_slot(
                        slot_id, (subtask.session_id, subtask.subtask_id)
                    )
                    logger.debug(
                        "Released slot %d for subtask %s", slot_id, subtask.subtask_id
                    )
                await quota_ref.release_quotas(tuple(batch_quota_req.keys()))

        # TODO(fyrestone): For the retryable op, we should rerun it when
        #  any exceptions occurred.
        if subtask.retryable:
            return await _retry_run(subtask, subtask_info, _run_subtask_once)
        else:
            try:
                return await _run_subtask_once()
            except Exception as e:
                unretryable_op = [
                    chunk.op
                    for chunk in subtask.chunk_graph
                    if not getattr(chunk.op, "retryable", True)
                ]
                message = (
                    f"Run subtask failed due to {e}, the subtask {subtask.subtask_id} is "
                    f"not retryable, it contains unretryable op: \n"
                    f"{pprint.pformat(unretryable_op)}"
                )
                logger.error(message)

                raise wrap_exception(
                    e, wrap_name="_UnretryableException", message=message
                )

    async def run_subtask(
        self, subtask: Subtask, band_name: str, supervisor_address: str
    ):
        if subtask.subtask_id in self._subtask_info:  # pragma: no cover
            raise Exception(
                f"Subtask {subtask.subtask_id} is already running on this band[{self.address}]."
            )
        logger.debug(
            "Start to schedule subtask %s on %s.", subtask.subtask_id, self.address
        )
        self._submitted_subtask_count.record(1, {"band": self.address})
        with mo.debug.no_message_trace():
            task = asyncio.create_task(
                self.ref().internal_run_subtask(subtask, band_name)
            )

        logger.debug("Subtask %r accepted in worker %s", subtask, self.address)
        # the extra_config may be None. the extra config overwrites the default value.
        subtask_max_retries = (
            subtask.extra_config.get("subtask_max_retries")
            if subtask.extra_config
            else None
        )
        if subtask_max_retries is None:
            subtask_max_retries = self._subtask_max_retries

        self._subtask_info[subtask.subtask_id] = SubtaskExecutionInfo(
            task, band_name, supervisor_address, max_retries=subtask_max_retries
        )
        result = await task
        self._subtask_info.pop(subtask.subtask_id, None)
        self._finished_subtask_count.record(1, {"band": self.address})
        logger.debug("Subtask %s finished with result %s", subtask.subtask_id, result)
        return result

    async def cancel_subtask(self, subtask_id: str, kill_timeout: Optional[int] = 5):
        try:
            subtask_info = self._subtask_info[subtask_id]
        except KeyError:
            logger.info("Subtask %s not exists, skip cancel.", subtask_id)
            return
        logger.info(
            "Start to cancel subtask %s in slot %s, kill_timeout is %s",
            subtask_id,
            subtask_info.slot_id,
            kill_timeout,
        )

        kill_timeout = kill_timeout if self._enable_kill_slot else None
        if not subtask_info.cancelling:
            subtask_info.kill_timeout = kill_timeout
            subtask_info.cancelling = True
            subtask_info.aio_task.cancel()

        await subtask_info.aio_task
        self._subtask_info.pop(subtask_id, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

from ...core import recursive_tile
from ...serialization.serializables import (
    FieldTypes,
    KeyField,
    StringField,
    Int32Field,
    Int64Field,
    TupleField,
)
from ...utils import has_unknown_shape
from ..utils import validate_axis, decide_chunk_sizes
from ..operands import TensorHasInput, TensorOperandMixin
from ..array_utils import get_array_module


class TensorFFTBaseMixin(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def _get_shape(cls, op, shape):
        raise NotImplementedError

    @classmethod
    def _tile_fft(cls, op, axes):
        in_tensor = op.inputs[0]
        out_tensor = op.outputs[0]

        if any(in_tensor.chunk_shape[axis] != 1 for axis in axes):
            if has_unknown_shape(in_tensor):
                yield
            # fft requires only 1 chunk for the specified axis, so we do rechunk first
            chunks = {
                validate_axis(in_tensor.ndim, axis): in_tensor.shape[axis]
                for axis in axes
            }
            new_chunks = decide_chunk_sizes(
                in_tensor.shape, chunks, in_tensor.dtype.itemsize
            )
            in_tensor = yield from recursive_tile(in_tensor.rechunk(new_chunks))

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk_shape = cls._get_shape(op, c.shape)
            out_chunk = chunk_op.new_chunk(
                [c], shape=chunk_shape, index=c.index, order=out_tensor.order
            )
            out_chunks.append(out_chunk)

        nsplits = [
            tuple(
                c.shape[i]
                for c in out_chunks
                if all(idx == 0 for j, idx in enumerate(c.index) if j != i)
            )
            for i in range(len(out_chunks[0].shape))
        ]
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            out_tensor.shape,
            order=out_tensor.order,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    def __call__(self, a, order=None):
        shape = self._get_shape(self, a.shape)
        order = a.order if order is None else order
        return self.new_tensor([a], shape, order=order)


class TensorFFTMixin(TensorFFTBaseMixin):
    __slots__ = ()

    @classmethod
    def tile(cls, op):
        return (yield from cls._tile_fft(op, [op.axis]))


class TensorComplexFFTMixin(TensorFFTMixin):
    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = list(shape)
        if op.n is not None:
            new_shape[op.axis] = op.n
        return tuple(new_shape)


def validate_fft(tensor, axis=-1, norm=None):
    validate_axis(tensor.ndim, axis)
    if norm is not None and norm not in ("ortho",):
        raise ValueError(f'Invalid norm value {norm}, should be None or "ortho"')


class TensorFFTNMixin(TensorFFTBaseMixin):
    @classmethod
    def tile(cls, op):
        return (yield from cls._tile_fft(op, op.axes))

    @staticmethod
    def _merge_shape(op, shape):
        new_shape = list(shape)
        if op.shape is not None:
            for ss, axis in zip(op.shape, op.axes):
                new_shape[axis] = ss
        return new_shape


class TensorComplexFFTNMixin(TensorFFTNMixin):
    @classmethod
    def _get_shape(cls, op, shape):
        return tuple(cls._merge_shape(op, shape))


class TensorRealFFTNMixin(TensorFFTNMixin):
    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = cls._merge_shape(op, shape)
        new_shape[op.axes[-1]] = new_shape[op.axes[-1]] // 2 + 1
        return tuple(new_shape)


class TensorRealIFFTNMixin(TensorFFTNMixin):
    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = list(shape)
        new_shape[op.axes[-1]] = 2 * (new_shape[op.axes[-1]] - 1)
        return tuple(cls._merge_shape(op, new_shape))


def validate_fftn(tensor, s=None, axes=None, norm=None):
    if axes is None:
        if s is None:
            axes = tuple(range(tensor.ndim))
        else:
            axes = tuple(range(len(s)))
    else:
        for axis in axes:
            validate_axis(tensor.ndim, axis)
        if len(set(axes)) < len(axes):
            raise ValueError("Duplicate axes not allowed")

    if norm is not None and norm not in ("ortho",):
        raise ValueError(f'Invalid norm value {norm}, should be None or "ortho"')

    return axes


class TensorFFTShiftMixin(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def _is_inverse(cls):
        return False

    @classmethod
    def _process_axes(cls, x, axes):
        if axes is None:
            axes = tuple(range(x.ndim))
        elif isinstance(axes, Iterable):
            axes = tuple(axes)
        else:
            axes = (axes,)

        return axes

    @classmethod
    def tile(cls, op):
        from ..merge import concatenate

        axes = op.axes
        in_tensor = op.input
        is_inverse = cls._is_inverse()

        if has_unknown_shape(in_tensor):
            yield

        x = in_tensor
        for axis in axes:
            size = in_tensor.shape[axis]
            slice_on = (size + 1) // 2 if not is_inverse else size // 2
            slc1 = [slice(None)] * axis + [slice(slice_on)]
            slc2 = [slice(None)] * axis + [slice(slice_on, None)]
            x = concatenate([x[slc2], x[slc1]], axis=axis)

        x = yield from recursive_tile(x)
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs, op.outputs[0].shape, chunks=x.chunks, nsplits=x.nsplits
        )


class TensorDiscreteFourierTransform(TensorHasInput):
    __slots__ = ()


class TensorBaseFFT(TensorDiscreteFourierTransform):
    _input = KeyField("input")
    _norm = StringField("norm")

    @property
    def norm(self):
        return getattr(self, "_norm", None)


class TensorBaseSingleDimensionFFT(TensorBaseFFT):
    _n = Int64Field("n")
    _axis = Int32Field("axis")

    @property
    def n(self):
        return self._n

    @property
    def axis(self):
        return self._axis

    @classmethod
    def execute(cls, ctx, op):
        a = ctx[op.inputs[0].key]
        xp = get_array_module(a)
        fun = _get_fft_func(op, xp)
        res = fun(a, n=op.n, axis=op.axis, norm=op.norm)
        if res.dtype != op.dtype:
            res = res.astype(op.dtype)
        ctx[op.outputs[0].key] = res


class TensorBaseMultipleDimensionFFT(TensorBaseFFT):
    _shape = TupleField("shape", FieldTypes.int64)
    _axes = TupleField("axes", FieldTypes.int32)

    @property
    def shape(self):
        return self._shape

    @property
    def axes(self):
        return self._axes

    @classmethod
    def execute(cls, ctx, op):
        a = ctx[op.inputs[0].key]
        xp = get_array_module(a)
        fun = _get_fft_func(op, xp)
        res = fun(a, s=op.shape, axes=op.axes, norm=op.norm)
        if res.dtype != op.dtype:
            res = res.astype(op.dtype)
        ctx[op.outputs[0].key] = res


def _get_fft_func(op, xp):
    fun_name = type(op).__name__.lower()[6:]  # all op starts with tensor
    return getattr(xp.fft, fun_name)


class TensorStandardFFT(TensorBaseSingleDimensionFFT):
    pass


class TensorStandardFFTN(TensorBaseMultipleDimensionFFT):
    pass


class TensorFFTShiftBase(TensorHasInput):
    _input = KeyField("input")
    _axes = TupleField("axes", FieldTypes.int32)

    @property
    def axes(self):
        return self._axes


class TensorRealFFT(TensorBaseSingleDimensionFFT):
    pass


class TensorRealFFTN(TensorBaseMultipleDimensionFFT):
    pass


class TensorHermitianFFT(TensorBaseSingleDimensionFFT):
    pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Union

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....utils import parse_readable_size
from ..core import SessionInfo
from ..supervisor import CustomLogMetaActor, SessionManagerActor, SessionActor
from ..worker import CustomLogActor
from .core import AbstractSessionAPI


class SessionAPI(AbstractSessionAPI):
    def __init__(
        self, address: str, session_manager: mo.ActorRefType[SessionManagerActor]
    ):
        self._address = address
        self._session_manager_ref = session_manager

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls, address: str, **kwargs) -> "SessionAPI":
        if kwargs:  # pragma: no cover
            raise TypeError(f"SessionAPI.create got unknown arguments: {list(kwargs)}")
        session_manager = await mo.actor_ref(address, SessionManagerActor.default_uid())
        return SessionAPI(address, session_manager)

    async def create_session(self, session_id: str) -> str:
        session_actor_ref = await self._session_manager_ref.create_session(session_id)
        return session_actor_ref.address

    async def get_sessions(self) -> List[SessionInfo]:
        return await self._session_manager_ref.get_sessions()

    async def has_session(self, session_id: str) -> bool:
        """
        Check if session created.

        Parameters
        ----------
        session_id : str
            Session ID.

        Returns
        -------
        if_exists : bool
        """
        return await self._session_manager_ref.has_session(session_id)

    async def delete_session(self, session_id: str):
        await self._session_manager_ref.delete_session(session_id)

    async def delete_all_sessions(self):
        await self._session_manager_ref.delete_all_sessions()

    @alru_cache(cache_exceptions=False)
    async def get_session_address(self, session_id: str) -> str:
        """
        Get session address.

        Parameters
        ----------
        session_id : str
            Session ID.

        Returns
        -------
        address : str
            Session address.
        """
        return (await self._session_manager_ref.get_session_ref(session_id)).address

    async def get_last_idle_time(
        self, session_id: Union[str, None] = None
    ) -> Union[float, None]:
        return await self._session_manager_ref.get_last_idle_time(session_id)

    @alru_cache(cache_exceptions=False)
    async def _get_session_ref(self, session_id: str) -> mo.ActorRefType[SessionActor]:
        return await self._session_manager_ref.get_session_ref(session_id)

    async def create_remote_object(
        self, session_id: str, name: str, object_cls, *args, **kwargs
    ):
        session = await self._get_session_ref(session_id)
        return await session.create_remote_object(name, object_cls, *args, **kwargs)

    async def get_remote_object(self, session_id: str, name: str):
        session = await self._get_session_ref(session_id)
        return await session.get_remote_object(name)

    async def destroy_remote_object(self, session_id: str, name: str):
        session = await self._get_session_ref(session_id)
        return await session.destroy_remote_object(name)

    @alru_cache(cache_exceptions=False)
    async def _get_custom_log_meta_ref(
        self, session_id: str
    ) -> mo.ActorRefType[CustomLogMetaActor]:
        session = await self._get_session_ref(session_id)
        return await mo.actor_ref(
            mo.ActorRef(session.address, CustomLogMetaActor.gen_uid(session_id))
        )

    async def register_custom_log_path(
        self,
        session_id: str,
        tileable_op_key: str,
        chunk_op_key: str,
        worker_address: str,
        log_path: str,
    ):
        custom_log_meta_ref = await self._get_custom_log_meta_ref(session_id)
        return await custom_log_meta_ref.register_custom_log_path(
            tileable_op_key, chunk_op_key, worker_address, log_path
        )

    @classmethod
    async def new_custom_log_dir(cls, address: str, session_id: str):
        try:
            ref = await mo.actor_ref(mo.ActorRef(address, CustomLogActor.default_uid()))
        except mo.ActorNotExist:
            return
        return await ref.new_custom_log_dir(session_id)

    async def fetch_tileable_op_logs(
        self,
        session_id: str,
        tileable_op_key: str,
        chunk_op_key_to_offsets: Dict[str, List[int]],
        chunk_op_key_to_sizes: Dict[str, List[int]],
    ) -> Dict:
        custom_log_meta_ref = await self._get_custom_log_meta_ref(session_id)
        chunk_op_key_to_arr_paths = await custom_log_meta_ref.get_tileable_op_log_paths(
            tileable_op_key
        )
        if chunk_op_key_to_arr_paths is None:
            return
        worker_to_kwds = dict()
        for chunk_op_key, addr_path in chunk_op_key_to_arr_paths.items():
            worker_address, log_path = addr_path
            if isinstance(chunk_op_key_to_offsets, dict):
                offset = chunk_op_key_to_offsets.get(chunk_op_key, 0)
            elif isinstance(chunk_op_key_to_offsets, str):
                offset = int(parse_readable_size(chunk_op_key_to_offsets)[0])
            elif isinstance(chunk_op_key_to_offsets, int):
                offset = chunk_op_key_to_offsets
            else:
                offset = 0
            if isinstance(chunk_op_key_to_sizes, dict):
                size = chunk_op_key_to_sizes.get(chunk_op_key, -1)
            elif isinstance(chunk_op_key_to_sizes, str):
                size = int(parse_readable_size(chunk_op_key_to_sizes)[0])
            elif isinstance(chunk_op_key_to_sizes, int):
                size = chunk_op_key_to_sizes
            else:
                size = -1
            if worker_address not in worker_to_kwds:
                worker_to_kwds[worker_address] = {
                    "chunk_op_keys": [],
                    "log_paths": [],
                    "offsets": [],
                    "sizes": [],
                }
            kwds = worker_to_kwds[worker_address]
            kwds["chunk_op_keys"].append(chunk_op_key)
            kwds["log_paths"].append(log_path)
            kwds["offsets"].append(offset)
            kwds["sizes"].append(size)
        result = dict()
        for worker, kwds in worker_to_kwds.items():
            custom_log_ref = await mo.actor_ref(
                mo.ActorRef(worker, CustomLogActor.default_uid())
            )
            chunk_op_keys = kwds.pop("chunk_op_keys")
            logs = await custom_log_ref.fetch_logs(**kwds)
            for chunk_op_key, log_result in zip(chunk_op_keys, logs):
                result[chunk_op_key] = log_result
        return result


class MockSessionAPI(SessionAPI):
    @classmethod
    async def create(cls, address: str, **kwargs) -> "SessionAPI":
        session_id = kwargs.pop("session_id")
        if kwargs:  # pragma: no cover
            raise TypeError(f"SessionAPI.create got unknown arguments: {list(kwargs)}")

        session_manager = await mo.create_actor(
            SessionManagerActor, address=address, uid=SessionManagerActor.default_uid()
        )
        if session_id:
            await session_manager.create_session(session_id, create_services=False)
        return MockSessionAPI(address, session_manager)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from typing import List, Set, Union, Generator

from ....typing import EntityType
from ..entity import EntityGraph, TileableGraph, ChunkGraph


def _default_inputs_selector(inputs: List[EntityType]) -> List[EntityType]:
    return inputs


class AbstractGraphBuilder(ABC):
    _graph: EntityGraph

    def __init__(self, graph: EntityGraph):
        self._graph = graph

    def _process_node(self, entity: EntityType):
        return entity

    def _select_inputs(self, inputs: List[EntityType]):
        return inputs

    def _if_add_node(
        self, node: EntityType, visited: Set
    ):  # pylint: disable=no-self-use
        return node not in visited

    def _add_nodes(
        self,
        graph: Union[ChunkGraph, TileableGraph],
        nodes: List[EntityType],
        visited: Set,
    ):
        # update visited
        visited.update(nodes)

        while len(nodes) > 0:
            node = nodes.pop()
            node = self._process_node(node)

            # mark node as visited
            visited.add(node)

            # add node to graph if possible
            if not graph.contains(node):
                graph.add_node(node)

            children = self._select_inputs(node.inputs or [])
            if children:
                node.inputs = children
            for c in children:
                c = self._process_node(c)
                if not graph.contains(c):
                    graph.add_node(c)
                if not graph.has_successor(c, node):
                    graph.add_edge(c, node)
                for out in c.op.outputs:
                    if self._if_add_node(out, visited):
                        nodes.append(out)

    @abstractmethod
    def build(self) -> Generator[Union[EntityGraph, ChunkGraph], None, None]:
        """
        Build a entity graph.

        Returns
        -------
        graph : EntityGraph
            Entity graph.
        """
